"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Hazard.
"""

__all__ = ['Hazard']

import copy
import datetime as dt
import itertools
import logging
import pathlib
import warnings

import geopandas as gpd
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pathos.pools import ProcessPool as Pool
import rasterio
from rasterio.features import rasterize
from rasterio.warp import reproject, Resampling, calculate_default_transform
from scipy import sparse

from climada.hazard.tag import Tag as TagHazard
from climada.hazard.centroids.centr import Centroids
import climada.util.plot as u_plot
import climada.util.checker as u_check
import climada.util.dates_times as u_dt
from climada import CONFIG
import climada.util.hdf5_handler as u_hdf5
import climada.util.coordinates as u_coord
from climada.util.constants import ONE_LAT_KM
from climada.util.coordinates import NEAREST_NEIGHBOR_THRESHOLD

LOGGER = logging.getLogger(__name__)

DEF_VAR_EXCEL = {'sheet_name': {'inten': 'hazard_intensity',
                                'freq': 'hazard_frequency'
                                },
                 'col_name': {'cen_id': 'centroid_id/event_id',
                              'even_id': 'event_id',
                              'even_dt': 'event_date',
                              'even_name': 'event_name',
                              'freq': 'frequency',
                              'orig': 'orig_event_flag'
                              },
                 'col_centroids': {'sheet_name': 'centroids',
                                   'col_name': {'cen_id': 'centroid_id',
                                                'lat': 'latitude',
                                                'lon': 'longitude'
                                                }
                                   }
                 }
"""Excel variable names"""

DEF_VAR_MAT = {'field_name': 'hazard',
               'var_name': {'per_id': 'peril_ID',
                            'even_id': 'event_ID',
                            'ev_name': 'name',
                            'freq': 'frequency',
                            'inten': 'intensity',
                            'unit': 'units',
                            'frac': 'fraction',
                            'comment': 'comment',
                            'datenum': 'datenum',
                            'orig': 'orig_event_flag'
                            },
               'var_cent': {'field_names': ['centroids', 'hazard'],
                            'var_name': {'cen_id': 'centroid_ID',
                                         'lat': 'lat',
                                         'lon': 'lon'
                                         }
                            }
               }
"""MATLAB variable names"""


class Hazard():
    """
    Contains events of some hazard type defined at centroids. Loads from
    files with format defined in FILE_EXT.

    Attributes
    ----------
    tag : TagHazard
        information about the source
    units : str
        units of the intensity
    centroids : Centroids
        centroids of the events
    event_id : np.array
        id (>0) of each event
    event_name : list(str)
        name of each event (default: event_id)
    date : np.array
        integer date corresponding to the proleptic
        Gregorian ordinal, where January 1 of year 1 has ordinal 1
        (ordinal format of datetime library)
    orig : np.array
        flags indicating historical events (True)
        or probabilistic (False)
    frequency : np.array
        frequency of each event in years
    intensity : sparse.csr_matrix
        intensity of the events at centroids
    fraction : sparse.csr_matrix
        fraction of affected exposures for each
        event at each centroid
    """
    intensity_thres = 10
    """Intensity threshold per hazard used to filter lower intensities. To be
    set for every hazard type"""

    vars_oblig = {'tag',
                  'units',
                  'centroids',
                  'event_id',
                  'frequency',
                  'intensity',
                  'fraction'
                  }
    """Name of the variables needed to compute the impact. Types: scalar, str,
    list, 1dim np.array of size num_events, scipy.sparse matrix of shape
    num_events x num_centroids, Centroids and Tag."""

    vars_def = {'date',
                'orig',
                'event_name'
                }
    """Name of the variables used in impact calculation whose value is
    descriptive and can therefore be set with default values. Types: scalar,
    string, list, 1dim np.array of size num_events.
    """

    vars_opt = set()
    """Name of the variables that aren't need to compute the impact. Types:
    scalar, string, list, 1dim np.array of size num_events."""

    def __init__(self, haz_type='', pool=None):
        """
        Initialize values.

        Parameters
        ----------
        haz_type : str, optional
            acronym of the hazard type (e.g. 'TC').
        pool : pathos.pool, optional
            Pool that will be used for parallel computation when applicable. Default: None

        Examples
        --------
        Fill hazard values by hand:

        >>> haz = Hazard('TC')
        >>> haz.intensity = sparse.csr_matrix(np.zeros((2, 2)))
        >>> ...

        Take hazard values from file:

        >>> haz = Hazard.from_mat(HAZ_DEMO_MAT, 'demo')

        """
        self.tag = TagHazard()
        self.tag.haz_type = haz_type
        self.units = ''
        self.centroids = Centroids()
        # following values are defined for each event
        self.event_id = np.array([], int)
        self.frequency = np.array([], float)
        self.event_name = list()
        self.date = np.array([], int)
        self.orig = np.array([], bool)
        # following values are defined for each event and centroid
        self.intensity = sparse.csr_matrix(np.empty((0, 0)))  # events x centroids
        self.fraction = sparse.csr_matrix(np.empty((0, 0)))  # events x centroids
        if pool:
            self.pool = pool
            LOGGER.info('Using %s CPUs.', self.pool.ncpus)
        else:
            self.pool = None

    def clear(self):
        """Reinitialize attributes (except the process Pool)."""
        for (var_name, var_val) in self.__dict__.items():
            if isinstance(var_val, np.ndarray) and var_val.ndim == 1:
                setattr(self, var_name, np.array([], dtype=var_val.dtype))
            elif isinstance(var_val, sparse.csr_matrix):
                setattr(self, var_name, sparse.csr_matrix(np.empty((0, 0))))
            elif not isinstance(var_val, Pool):
                setattr(self, var_name, var_val.__class__())

    def check(self):
        """Check dimension of attributes.

        Raises
        ------
        ValueError
        """
        self.centroids.check()
        self._check_events()

    @classmethod
    def from_raster(cls, files_intensity, files_fraction=None, attrs=None,
                    band=None, haz_type=None, pool=None, src_crs=None, window=False,
                    geometry=False, dst_crs=False, transform=None, width=None,
                    height=None, resampling=Resampling.nearest):
        """Create Hazard with intensity and fraction values from raster files

        If raster files are masked, the masked values are set to 0.

        Files can be partially read using either window or geometry. Additionally, the data is
        reprojected when custom dst_crs and/or transform, width and height are specified.

        Parameters
        ----------
        files_intensity : list(str)
            file names containing intensity
        files_fraction : list(str)
            file names containing fraction
        attrs : dict, optional
            name of Hazard attributes and their values
        band : list(int), optional
            bands to read (starting at 1), default [1]
        haz_type : str, optional
            acronym of the hazard type (e.g. 'TC').
            Default: None, which will use the class default ('' for vanilla
            `Hazard` objects, and hard coded in some subclasses)
        pool : pathos.pool, optional
            Pool that will be used for parallel computation when applicable.
            Default: None
        src_crs : crs, optional
            source CRS. Provide it if error without it.
        window : rasterio.windows.Windows, optional
            window where data is
            extracted
        geometry : shapely.geometry, optional
            consider pixels only in shape
        dst_crs : crs, optional
            reproject to given crs
        transform : rasterio.Affine
            affine transformation to apply
        wdith : float, optional
            number of lons for transform
        height : float, optional
            number of lats for transform
        resampling : rasterio.warp.Resampling, optional
            resampling function used for reprojection to dst_crs

        Return
        ------
        Hazard
        """
        if isinstance(files_intensity, (str, pathlib.Path)):
            files_intensity = [files_intensity]
        if isinstance(files_fraction, (str, pathlib.Path)):
            files_fraction = [files_fraction]
        if not attrs:
            attrs = {}
        if not band:
            band = [1]
        if files_fraction is not None and len(files_intensity) != len(files_fraction):
            raise ValueError('Number of intensity files differs from fraction files: %s != %s'
                             % (len(files_intensity), len(files_fraction)))

        if haz_type is not None:
            try:
                haz = cls(haz_type=haz_type, pool=pool)
            except TypeError:
                haz = cls(pool=pool)
                LOGGER.warn("haz_type argument ('%s') is ignored, haz_type is determined by class",
                            haz_type)
        else:
            haz = cls(pool=pool)

        haz.tag.file_name = str(files_intensity) + ' ; ' + str(files_fraction)

        haz.centroids = Centroids.from_raster_file(
            files_intensity[0], src_crs=src_crs, window=window, geometry=geometry, dst_crs=dst_crs,
            transform=transform, width=width, height=height, resampling=resampling)
        if haz.pool:
            chunksize = min(len(files_intensity) // haz.pool.ncpus, 1000)
            inten_list = haz.pool.map(
                haz.centroids.values_from_raster_files,
                [[f] for f in files_intensity],
                itertools.repeat(band), itertools.repeat(src_crs),
                itertools.repeat(window), itertools.repeat(geometry),
                itertools.repeat(dst_crs), itertools.repeat(transform),
                itertools.repeat(width), itertools.repeat(height),
                itertools.repeat(resampling), chunksize=chunksize)
            haz.intensity = sparse.vstack(inten_list, format='csr')
            if files_fraction is not None:
                fract_list = haz.pool.map(
                    haz.centroids.values_from_raster_files,
                    [[f] for f in files_fraction],
                    itertools.repeat(band), itertools.repeat(src_crs),
                    itertools.repeat(window), itertools.repeat(geometry),
                    itertools.repeat(dst_crs), itertools.repeat(transform),
                    itertools.repeat(width), itertools.repeat(height),
                    itertools.repeat(resampling), chunksize=chunksize)
                haz.fraction = sparse.vstack(fract_list, format='csr')
        else:
            haz.intensity = haz.centroids.values_from_raster_files(
                files_intensity, band=band, src_crs=src_crs, window=window, geometry=geometry,
                dst_crs=dst_crs, transform=transform, width=width, height=height,
                resampling=resampling)
            if files_fraction is not None:
                haz.fraction = haz.centroids.values_from_raster_files(
                    files_fraction, band=band, src_crs=src_crs, window=window, geometry=geometry,
                    dst_crs=dst_crs, transform=transform, width=width, height=height,
                    resampling=resampling)

        if files_fraction is None:
            haz.fraction = haz.intensity.copy()
            haz.fraction.data.fill(1)

        if 'event_id' in attrs:
            haz.event_id = attrs['event_id']
        else:
            haz.event_id = np.arange(1, haz.intensity.shape[0] + 1)
        if 'frequency' in attrs:
            haz.frequency = attrs['frequency']
        else:
            haz.frequency = np.ones(haz.event_id.size)
        if 'event_name' in attrs:
            haz.event_name = attrs['event_name']
        else:
            haz.event_name = list(map(str, haz.event_id))
        if 'date' in attrs:
            haz.date = np.array([attrs['date']])
        else:
            haz.date = np.ones(haz.event_id.size)
        if 'orig' in attrs:
            haz.orig = np.array([attrs['orig']])
        else:
            haz.orig = np.ones(haz.event_id.size, bool)
        if 'unit' in attrs:
            haz.unit = attrs['unit']

        return haz

    def set_raster(self, *args, **kwargs):
        """This function is deprecated, use Hazard.from_raster."""
        LOGGER.warning("The use of Hazard.set_raster is deprecated."
                       "Use Hazard.from_raster instead.")
        self.__dict__ = Hazard.from_raster(*args, **kwargs).__dict__

    def set_vector(self, *args, **kwargs):
        """This function is deprecated, use Hazard.from_vector."""
        LOGGER.warning("The use of Hazard.set_vector is deprecated."
                       "Use Hazard.from_vector instead.")
        self.__dict__ = Hazard.from_vector(*args, **kwargs).__dict__

    @classmethod
    def from_vector(cls, files_intensity, files_fraction=None, attrs=None,
                    inten_name=None, frac_name=None, dst_crs=None, haz_type=None):
        """Read vector files format supported by fiona. Each intensity name is
        considered an event.

        Parameters
        ----------
        files_intensity : list(str)
            file names containing intensity, default: ['intensity']
        files_fraction : (list(str))
            file names containing fraction,
            default: ['fraction']
        attrs : dict, optional
            name of Hazard attributes and their values
        inten_name : list(str), optional
            name of variables containing the intensities of each event
        frac_name : list(str), optional
            name of variables containing
            the fractions of each event
        dst_crs : crs, optional
            reproject to given crs
        haz_type : str, optional
            acronym of the hazard type (e.g. 'TC').
            default: None, which will use the class default ('' for vanilla
            `Hazard` objects, hard coded in some subclasses)

        Returns
        -------
        haz : climada.hazard.Hazard
            Hazard from vector file
        """
        if not attrs:
            attrs = {}
        if not inten_name:
            inten_name = ['intensity']
        if not frac_name:
            inten_name = ['fraction']
        if files_fraction is not None and len(files_intensity) != len(files_fraction):
            raise ValueError('Number of intensity files differs from fraction files: %s != %s'
                             % (len(files_intensity), len(files_fraction)))

        haz = cls() if haz_type is None else cls(haz_type)
        haz.tag.file_name = str(files_intensity) + ' ; ' + str(files_fraction)

        if len(files_intensity) > 0:
            haz.centroids = Centroids.from_vector_file(files_intensity[0], dst_crs=dst_crs)
        elif files_fraction is not None and len(files_fraction) > 0:
            haz.centroids = Centroids.from_vector_file(files_fraction[0], dst_crs=dst_crs)
        else:
            haz.centroids = Centroids()

        haz.intensity = haz.centroids.values_from_vector_files(
            files_intensity, val_names=inten_name, dst_crs=dst_crs)
        if files_fraction is None:
            haz.fraction = haz.intensity.copy()
            haz.fraction.data.fill(1)
        else:
            haz.fraction = haz.centroids.values_from_vector_files(
                files_fraction, val_names=frac_name, dst_crs=dst_crs)

        if 'event_id' in attrs:
            haz.event_id = attrs['event_id']
        else:
            haz.event_id = np.arange(1, haz.intensity.shape[0] + 1)
        if 'frequency' in attrs:
            haz.frequency = attrs['frequency']
        else:
            haz.frequency = np.ones(haz.event_id.size)
        if 'event_name' in attrs:
            haz.event_name = attrs['event_name']
        else:
            haz.event_name = list(map(str, haz.event_id))
        if 'date' in attrs:
            haz.date = np.array([attrs['date']])
        else:
            haz.date = np.ones(haz.event_id.size)
        if 'orig' in attrs:
            haz.orig = np.array([attrs['orig']])
        else:
            haz.orig = np.ones(haz.event_id.size, bool)
        if 'unit' in attrs:
            haz.unit = attrs['unit']

        return haz

    def reproject_raster(self, dst_crs=False, transform=None, width=None, height=None,
                         resampl_inten=Resampling.nearest, resampl_fract=Resampling.nearest):
        """Change current raster data to other CRS and/or transformation

        Parameters
        ----------
        dst_crs: crs, optional
            reproject to given crs
        transform: rasterio.Affine
            affine transformation to apply
        wdith: float
            number of lons for transform
        height: float
            number of lats for transform
        resampl_inten: rasterio.warp,.Resampling optional
            resampling function used for reprojection to dst_crs for intensity
        resampl_fract: rasterio.warp,.Resampling optional
            resampling function used for reprojection to dst_crs for fraction
        """
        if not self.centroids.meta:
            raise ValueError('Raster not set')
        if not dst_crs:
            dst_crs = self.centroids.meta['crs']
        if transform and not width or transform and not height:
            raise ValueError('Provide width and height to given transformation.')
        if not transform:
            transform, width, height = calculate_default_transform(
                self.centroids.meta['crs'], dst_crs, self.centroids.meta['width'],
                self.centroids.meta['height'], self.centroids.meta['transform'][2],
                (self.centroids.meta['transform'][5]
                 + self.centroids.meta['height'] * self.centroids.meta['transform'][4]),
                (self.centroids.meta['transform'][2]
                 + self.centroids.meta['width'] * self.centroids.meta['transform'][0]),
                self.centroids.meta['transform'][5])
        dst_meta = self.centroids.meta.copy()
        dst_meta.update({'crs': dst_crs, 'transform': transform,
                         'width': width, 'height': height
                         })
        intensity = np.zeros((self.size, dst_meta['height'], dst_meta['width']))
        fraction = np.zeros((self.size, dst_meta['height'], dst_meta['width']))
        kwargs = {'src_transform': self.centroids.meta['transform'],
                  'src_crs': self.centroids.meta['crs'],
                  'dst_transform': transform, 'dst_crs': dst_crs,
                  'resampling': resampl_inten}
        for idx_ev, inten in enumerate(self.intensity.toarray()):
            reproject(
                source=np.asarray(inten.reshape((self.centroids.meta['height'],
                                                 self.centroids.meta['width']))),
                destination=intensity[idx_ev, :, :],
                **kwargs)
        kwargs.update(resampling=resampl_fract)
        for idx_ev, fract in enumerate(self.fraction.toarray()):
            reproject(
                source=np.asarray(
                    fract.reshape((self.centroids.meta['height'],
                                   self.centroids.meta['width']))),
                destination=fraction[idx_ev, :, :],
                **kwargs)
        self.centroids.meta = dst_meta
        self.intensity = sparse.csr_matrix(
            intensity.reshape(self.size, dst_meta['height'] * dst_meta['width']))
        self.fraction = sparse.csr_matrix(
            fraction.reshape(self.size, dst_meta['height'] * dst_meta['width']))
        self.check()

    def reproject_vector(self, dst_crs, scheduler=None):
        """Change current point data to a a given projection

        Parameters
        ----------
        dst_crs : crs
            reproject to given crs
        scheduler : str, optional
            used for dask map_partitions. “threads”,
            “synchronous” or “processes”
        """
        self.centroids.set_geometry_points(scheduler)
        self.centroids.geometry = self.centroids.geometry.to_crs(dst_crs)
        self.centroids.lat = self.centroids.geometry[:].y
        self.centroids.lon = self.centroids.geometry[:].x
        self.check()

    def raster_to_vector(self):
        """Change current raster to points (center of the pixels)"""
        self.centroids.set_meta_to_lat_lon()
        self.centroids.meta = dict()
        self.check()

    def vector_to_raster(self, scheduler=None):
        """Change current point data to a raster with same resolution

        Parameters
        ----------
        scheduler : str, optional
            used for dask map_partitions. “threads”,
            “synchronous” or “processes”
        """
        points_df = gpd.GeoDataFrame()
        points_df['latitude'] = self.centroids.lat
        points_df['longitude'] = self.centroids.lon
        val_names = ['val' + str(i_ev) for i_ev in range(2 * self.size)]
        for i_ev, inten_name in enumerate(val_names):
            if i_ev < self.size:
                points_df[inten_name] = np.asarray(self.intensity[i_ev, :].toarray()).reshape(-1)
            else:
                points_df[inten_name] = np.asarray(self.fraction[i_ev - self.size, :].toarray()). \
                    reshape(-1)
        raster, meta = u_coord.points_to_raster(points_df, val_names,
                                                crs=self.centroids.geometry.crs,
                                                scheduler=scheduler)
        self.intensity = sparse.csr_matrix(raster[:self.size, :, :].reshape(self.size, -1))
        self.fraction = sparse.csr_matrix(raster[self.size:, :, :].reshape(self.size, -1))
        self.centroids = Centroids()
        self.centroids.meta = meta
        self.check()

    def read_mat(self, *args, **kwargs):
        """This function is deprecated, use Hazard.from_mat."""
        LOGGER.warning("The use of Hazard.read_mat is deprecated."
                       "Use Hazard.from_mat instead.")
        self.__dict__ = Hazard.from_mat(*args, **kwargs).__dict__

    @classmethod
    def from_mat(cls, file_name, description='', var_names=None):
        """Read climada hazard generate with the MATLAB code in .mat format.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names : dict, optional
            name of the variables in the file,
            default: DEF_VAR_MAT constant

        Returns
        -------
        haz : climada.hazard.Hazard
            Hazard object from the provided MATLAB file

        Raises
        ------
        KeyError
        """
        if not var_names:
            var_names = DEF_VAR_MAT
        LOGGER.info('Reading %s', file_name)
        haz = cls()
        haz.tag.file_name = str(file_name)
        haz.tag.description = description
        try:
            data = u_hdf5.read(file_name)
            try:
                data = data[var_names['field_name']]
            except KeyError:
                pass

            haz_type = u_hdf5.get_string(data[var_names['var_name']['per_id']])
            haz.tag.haz_type = haz_type
            haz.centroids = Centroids.from_mat(file_name, var_names=var_names['var_cent'])
            haz._read_att_mat(data, file_name, var_names)
        except KeyError as var_err:
            raise KeyError("Variable not in MAT file: " + str(var_err)) from var_err
        return haz

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use Hazard.from_excel."""
        LOGGER.warning("The use of Hazard.read_excel is deprecated."
                       "Use Hazard.from_excel instead.")
        self.__dict__ = Hazard.from_excel(*args, **kwargs).__dict__

    @classmethod
    def from_excel(cls, file_name, description='', var_names=None, haz_type=None):
        """Read climada hazard generated with the MATLAB code in Excel format.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names (dict, default): name of the variables in the file,
            default: DEF_VAR_EXCEL constant
        haz_type : str, optional
            acronym of the hazard type (e.g. 'TC').
            Default: None, which will use the class default ('' for vanilla `Hazard` objects, and hard coded in some
            subclasses)

        Returns
        -------
        haz : climada.hazard.Hazard
            Hazard object from the provided Excel file

        Raises
        ------
        KeyError
        """
        if not var_names:
            var_names = DEF_VAR_EXCEL
        LOGGER.info('Reading %s', file_name)
        haz = cls() if haz_type is None else cls(haz_type)
        haz.tag.file_name = file_name
        haz.tag.description = description
        try:
            haz.centroids = Centroids.from_excel(file_name, var_names=var_names['col_centroids'])
            haz._read_att_excel(file_name, var_names)
        except KeyError as var_err:
            raise KeyError("Variable not in Excel file: " + str(var_err)) from var_err
        return haz

    def select(self, event_names=None, date=None, orig=None, reg_id=None,
               extent=None, reset_frequency=False):
        """Select events matching provided criteria

        The frequency of events may need to be recomputed (see `reset_frequency`)!

        Parameters
        ----------
        event_names : list of str, optional
            Names of events.
        date : array-like of length 2 containing str or int, optional
            (initial date, final date) in string ISO format ('2011-01-02') or datetime
            ordinal integer.
        orig : bool, optional
            Select only historical (True) or only synthetic (False) events.
        reg_id : int, optional
            Region identifier of the centroids' region_id attibute.
        extent: tuple(float, float, float, float), optional
            Extent of centroids as (min_lon, max_lon, min_lat, max_lat).
            The default is None.
        reset_frequency : bool, optional
            Change frequency of events proportional to difference between first and last
            year (old and new). Default: False.

        Returns
        -------
        haz : Hazard or None
            If no event matching the specified criteria is found, None is returned.
        """
        if type(self) is Hazard:
            haz = Hazard(self.tag.haz_type)
        else:
            haz = self.__class__()

        #filter events
        sel_ev = np.ones(self.event_id.size, dtype=bool)

        # filter events by date
        if date is not None:
            date_ini, date_end = date
            if isinstance(date_ini, str):
                date_ini = u_dt.str_to_date(date[0])
                date_end = u_dt.str_to_date(date[1])
            sel_ev &= (date_ini <= self.date) & (self.date <= date_end)
            if not np.any(sel_ev):
                LOGGER.info('No hazard in date range %s.', date)
                return None

        # filter events hist/synthetic
        if isinstance(orig, bool):
            sel_ev &= (self.orig.astype(bool) == orig)
            if not np.any(sel_ev):
                LOGGER.info('No hazard with %s tracks.', str(orig))
                return None

        # filter events based on name
        sel_ev = np.argwhere(sel_ev).reshape(-1)
        if isinstance(event_names, list):
            filtered_events = [self.event_name[i] for i in sel_ev]
            try:
                new_sel = [filtered_events.index(n) for n in event_names]
            except ValueError as err:
                name = str(err).replace(" is not in list", "")
                LOGGER.info('No hazard with name %s', name)
                return None
            sel_ev = sel_ev[new_sel]

        # filter centroids
        sel_cen = self.centroids.select_mask(reg_id=reg_id, extent=extent)
        if not np.any(sel_cen):
            LOGGER.info('No hazard centroids within extent and region')
            return None

        sel_cen = sel_cen.nonzero()[0]
        for (var_name, var_val) in self.__dict__.items():
            if isinstance(var_val, np.ndarray) and var_val.ndim == 1 \
                    and var_val.size > 0:
                setattr(haz, var_name, var_val[sel_ev])
            elif isinstance(var_val, sparse.csr_matrix):
                setattr(haz, var_name, var_val[sel_ev, :][:, sel_cen])
            elif isinstance(var_val, list) and var_val:
                setattr(haz, var_name, [var_val[idx] for idx in sel_ev])
            elif var_name == 'centroids':
                if reg_id is None and extent is None:
                    new_cent = var_val
                else:
                    new_cent = var_val.select(sel_cen=sel_cen)
                setattr(haz, var_name, new_cent)
            else:
                setattr(haz, var_name, var_val)

        # reset frequency if date span has changed (optional):
        if reset_frequency:
            year_span_old = np.abs(dt.datetime.fromordinal(self.date.max()).year -
                                   dt.datetime.fromordinal(self.date.min()).year) + 1
            year_span_new = np.abs(dt.datetime.fromordinal(haz.date.max()).year -
                                   dt.datetime.fromordinal(haz.date.min()).year) + 1
            haz.frequency = haz.frequency * year_span_old / year_span_new

        haz.sanitize_event_ids()
        return haz

    def select_tight(self, buffer=NEAREST_NEIGHBOR_THRESHOLD/ONE_LAT_KM,
                     val='intensity'):
        """
        Reduce hazard to those centroids spanning a minimal box which
        contains all non-zero intensity or fraction points.

        Parameters
        ----------
        buffer : float, optional
            Buffer of box in the units of the centroids.
            The default is approximately equal to the default threshold
            from the assign_centroids method (works if centroids in
            lat/lon)
        val: string, optional
            Select tight by non-zero 'intensity' or 'fraction'. The
            default is 'intensity'.

        Returns
        -------
        Hazard
            Copy of the Hazard with centroids reduced to minimal box. All other
            hazard properties are carried over without changes.

        See also
        --------
        self.select: Method to select centroids by lat/lon extent
        util.coordinates.assign_coordinates: algorithm to match centroids.

        """

        if val == 'intensity':
            cent_nz = (self.intensity != 0).sum(axis=0).nonzero()[1]
        if val == 'fraction':
            cent_nz = (self.fraction != 0).sum(axis=0).nonzero()[1]
        lon_nz = self.centroids.lon[cent_nz]
        lat_nz = self.centroids.lat[cent_nz]
        ext = u_coord.latlon_bounds(lat=lat_nz, lon=lon_nz, buffer=buffer)
        return self.select(extent=(ext[0], ext[2], ext[1], ext[3]))

    def local_exceedance_inten(self, return_periods=(25, 50, 100, 250)):
        """Compute exceedance intensity map for given return periods.

        Parameters
        ----------
        return_periods : np.array
            return periods to consider

        Returns
        -------
        inten_stats: np.array
        """
        # warn if return period is above return period of rarest event:
        for period in return_periods:
            if period > 1 / self.frequency.min():
                LOGGER.warning('Return period %1.1f exceeds max. event return period.', period)
        LOGGER.info('Computing exceedance intenstiy map for return periods: %s',
                    return_periods)
        num_cen = self.intensity.shape[1]
        inten_stats = np.zeros((len(return_periods), num_cen))
        cen_step = CONFIG.max_matrix_size.int() // self.intensity.shape[0]
        if not cen_step:
            raise ValueError('Increase max_matrix_size configuration parameter to > %s'
                             % str(self.intensity.shape[0]))
        # separte in chunks
        chk = -1
        for chk in range(int(num_cen / cen_step)):
            self._loc_return_inten(
                np.array(return_periods),
                self.intensity[:, chk * cen_step:(chk + 1) * cen_step].toarray(),
                inten_stats[:, chk * cen_step:(chk + 1) * cen_step])
        self._loc_return_inten(
            np.array(return_periods),
            self.intensity[:, (chk + 1) * cen_step:].toarray(),
            inten_stats[:, (chk + 1) * cen_step:])
        # set values below 0 to zero if minimum of hazard.intensity >= 0:
        if self.intensity.min() >= 0 and np.min(inten_stats) < 0:
            LOGGER.warning('Exceedance intenstiy values below 0 are set to 0. \
                   Reason: no negative intensity values were found in hazard.')
            inten_stats[inten_stats < 0] = 0
        return inten_stats

    def plot_rp_intensity(self, return_periods=(25, 50, 100, 250),
                          smooth=True, axis=None, figsize=(9, 13), adapt_fontsize=True,
                          **kwargs):
        """Compute and plot hazard exceedance intensity maps for different
        return periods. Calls local_exceedance_inten.

        Parameters
        ----------
        return_periods: tuple(int), optional
            return periods to consider
        smooth: bool, optional
            smooth plot to plot.RESOLUTIONxplot.RESOLUTION
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        figsize: tuple, optional
            figure size for plt.subplots
        kwargs: optional
            arguments for pcolormesh matplotlib function used in event plots

        Returns
        -------
        axis, inten_stats:  matplotlib.axes._subplots.AxesSubplot, np.ndarray
            intenstats is return_periods.size x num_centroids
        """
        self._set_coords_centroids()
        inten_stats = self.local_exceedance_inten(np.array(return_periods))
        colbar_name = 'Intensity (' + self.units + ')'
        title = list()
        for ret in return_periods:
            title.append('Return period: ' + str(ret) + ' years')
        axis = u_plot.geo_im_from_array(inten_stats, self.centroids.coord,
                                        colbar_name, title, smooth=smooth,
                                        axes=axis, figsize=figsize, adapt_fontsize=adapt_fontsize, **kwargs)
        return axis, inten_stats

    def plot_intensity(self, event=None, centr=None, smooth=True, axis=None, adapt_fontsize=True,
                       **kwargs):
        """Plot intensity values for a selected event or centroid.

        Parameters
        ----------
        event: int or str, optional
            If event > 0, plot intensities of
            event with id = event. If event = 0, plot maximum intensity in
            each centroid. If event < 0, plot abs(event)-largest event. If
            event is string, plot events with that name.
        centr: int or tuple, optional
            If centr > 0, plot intensity
            of all events at centroid with id = centr. If centr = 0,
            plot maximum intensity of each event. If centr < 0,
            plot abs(centr)-largest centroid where higher intensities
            are reached. If tuple with (lat, lon) plot intensity of nearest
            centroid.
        smooth: bool, optional
            Rescale data to RESOLUTIONxRESOLUTION pixels (see constant
            in module `climada.util.plot`)
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs: optional
            arguments for pcolormesh matplotlib function
            used in event plots or for plot function used in centroids plots

        Returns
        -------
            matplotlib.axes._subplots.AxesSubplot

        Raises
        ------
            ValueError
        """
        self._set_coords_centroids()
        col_label = 'Intensity (%s)' % self.units
        crs_epsg, _ = u_plot.get_transformation(self.centroids.geometry.crs)
        if event is not None:
            if isinstance(event, str):
                event = self.get_event_id(event)
            return self._event_plot(event, self.intensity, col_label,
                                    smooth, crs_epsg, axis, adapt_fontsize=adapt_fontsize, **kwargs)
        if centr is not None:
            if isinstance(centr, tuple):
                _, _, centr = self.centroids.get_closest_point(centr[0], centr[1])
            return self._centr_plot(centr, self.intensity, col_label, axis, **kwargs)

        raise ValueError("Provide one event id or one centroid id.")

    def plot_fraction(self, event=None, centr=None, smooth=True, axis=None,
                      **kwargs):
        """Plot fraction values for a selected event or centroid.

        Parameters
        ----------
        event: int or str, optional
            If event > 0, plot fraction of event
            with id = event. If event = 0, plot maximum fraction in each
            centroid. If event < 0, plot abs(event)-largest event. If event
            is string, plot events with that name.
        centr: int or tuple, optional
            If centr > 0, plot fraction
            of all events at centroid with id = centr. If centr = 0,
            plot maximum fraction of each event. If centr < 0,
            plot abs(centr)-largest centroid where highest fractions
            are reached. If tuple with (lat, lon) plot fraction of nearest
            centroid.
        smooth: bool, optional
            Rescale data to RESOLUTIONxRESOLUTION pixels (see constant
            in module `climada.util.plot`)
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs: optional
            arguments for pcolormesh matplotlib function
            used in event plots or for plot function used in centroids plots

        Returns
        -------
            matplotlib.axes._subplots.AxesSubplot

        Raises
        ------
            ValueError
        """
        self._set_coords_centroids()
        col_label = 'Fraction'
        if event is not None:
            if isinstance(event, str):
                event = self.get_event_id(event)
            return self._event_plot(event, self.fraction, col_label, smooth, axis,
                                    **kwargs)
        if centr is not None:
            if isinstance(centr, tuple):
                _, _, centr = self.centroids.get_closest_point(centr[0], centr[1])
            return self._centr_plot(centr, self.fraction, col_label, axis, **kwargs)

        raise ValueError("Provide one event id or one centroid id.")

    def sanitize_event_ids(self):
        """Make sure that event ids are unique"""
        if np.unique(self.event_id).size != self.event_id.size:
            LOGGER.debug('Resetting event_id.')
            self.event_id = np.arange(1, self.event_id.size + 1)

    def get_event_id(self, event_name):
        """Get an event id from its name. Several events might have the same
        name.

        Parameters
        ----------
        event_name: str
            Event name

        Returns
        -------
        list_id: np.array(int)
        """
        list_id = self.event_id[[i_name for i_name, val_name in enumerate(self.event_name)
                                 if val_name == event_name]]
        if list_id.size == 0:
            raise ValueError("No event with name: %s" % event_name)
        return list_id

    def get_event_name(self, event_id):
        """Get the name of an event id.

        Parameters
        ----------
        event_id: int
            id of the event

        Returns
        -------
            str

        Raises
        ------
            ValueError
        """
        try:
            return self.event_name[np.argwhere(
                self.event_id == event_id)[0][0]]
        except IndexError as err:
            raise ValueError(f"No event with id: {event_id}") from err

    def get_event_date(self, event=None):
        """Return list of date strings for given event or for all events,
        if no event provided.

        Parameters
        ----------
        event: str or int, optional
            event name or id.

        Returns
        -------
        l_dates: list(str)
        """
        if event is None:
            l_dates = [u_dt.date_to_str(date) for date in self.date]
        elif isinstance(event, str):
            ev_ids = self.get_event_id(event)
            l_dates = [
                u_dt.date_to_str(self.date[np.argwhere(self.event_id == ev_id)[0][0]])
                for ev_id in ev_ids]
        else:
            ev_idx = np.argwhere(self.event_id == event)[0][0]
            l_dates = [u_dt.date_to_str(self.date[ev_idx])]
        return l_dates

    def calc_year_set(self):
        """From the dates of the original events, get number yearly events.

        Returns
        -------
        orig_yearset: dict
            key are years, values array with event_ids of that year

        """
        orig_year = np.array([dt.datetime.fromordinal(date).year
                              for date in self.date[self.orig]])
        orig_yearset = {}
        for year in np.unique(orig_year):
            orig_yearset[year] = self.event_id[self.orig][orig_year == year]
        return orig_yearset

    def remove_duplicates(self):
        """Remove duplicate events (events with same name and date)."""
        events = list(zip(self.event_name, self.date))
        set_ev = set(events)
        if len(set_ev) == self.event_id.size:
            return
        unique_pos = sorted([events.index(event) for event in set_ev])
        for var_name, var_val in vars(self).items():
            if isinstance(var_val, sparse.csr.csr_matrix):
                setattr(self, var_name, var_val[unique_pos, :])
            elif isinstance(var_val, np.ndarray) and var_val.ndim == 1:
                setattr(self, var_name, var_val[unique_pos])
            elif isinstance(var_val, list) and len(var_val) > 0:
                setattr(self, var_name, [var_val[p] for p in unique_pos])

    def set_frequency(self, yearrange=None):
        """Set hazard frequency from yearrange or intensity matrix.

        Parameters
        ----------
        yearrange: tuple or list, optional
            year range to be used to compute frequency
            per event. If yearrange is not given (None), the year range is
            derived from self.date
        """
        if not yearrange:
            delta_time = dt.datetime.fromordinal(int(np.max(self.date))).year - \
                         dt.datetime.fromordinal(int(np.min(self.date))).year + 1
        else:
            delta_time = max(yearrange) - min(yearrange) + 1
        num_orig = self.orig.nonzero()[0].size
        if num_orig > 0:
            ens_size = self.event_id.size / num_orig
        else:
            ens_size = 1
        self.frequency = np.ones(self.event_id.size) / delta_time / ens_size

    @property
    def size(self):
        """Return number of events."""
        return self.event_id.size

    def write_raster(self, file_name, intensity=True):
        """Write intensity or fraction as GeoTIFF file. Each band is an event

        Parameters
        ----------
        file_name: str
            file name to write in tif format
        intensity: bool
            if True, write intensity, otherwise write fraction
        """
        variable = self.intensity
        if not intensity:
            variable = self.fraction
        if self.centroids.meta:
            u_coord.write_raster(file_name, variable.toarray(), self.centroids.meta)
        else:
            pixel_geom = self.centroids.calc_pixels_polygons()
            profile = self.centroids.meta
            profile.update(driver='GTiff', dtype=rasterio.float32, count=self.size)
            with rasterio.open(file_name, 'w', **profile) as dst:
                LOGGER.info('Writing %s', file_name)
                for i_ev in range(variable.shape[0]):
                    raster = rasterize(
                        [(x, val) for (x, val) in
                         zip(pixel_geom, np.array(variable[i_ev, :].toarray()).reshape(-1))],
                        out_shape=(profile['height'], profile['width']),
                        transform=profile['transform'], fill=0,
                        all_touched=True, dtype=profile['dtype'], )
                    dst.write(raster.astype(profile['dtype']), i_ev + 1)

    def write_hdf5(self, file_name, todense=False):
        """Write hazard in hdf5 format.

        Parameters
        ----------
        file_name: str
            file name to write, with h5 format
        """
        LOGGER.info('Writing %s', file_name)
        hf_data = h5py.File(file_name, 'w')
        str_dt = h5py.special_dtype(vlen=str)
        for (var_name, var_val) in self.__dict__.items():
            if var_name == 'centroids':
                self.centroids.write_hdf5(hf_data.create_group(var_name))
            elif var_name == 'tag':
                hf_str = hf_data.create_dataset('haz_type', (1,), dtype=str_dt)
                hf_str[0] = var_val.haz_type
                hf_str = hf_data.create_dataset('file_name', (1,), dtype=str_dt)
                hf_str[0] = str(var_val.file_name)
                hf_str = hf_data.create_dataset('description', (1,), dtype=str_dt)
                hf_str[0] = str(var_val.description)
            elif isinstance(var_val, sparse.csr_matrix):
                if todense:
                    hf_data.create_dataset(var_name, data=var_val.toarray())
                else:
                    hf_csr = hf_data.create_group(var_name)
                    hf_csr.create_dataset('data', data=var_val.data)
                    hf_csr.create_dataset('indices', data=var_val.indices)
                    hf_csr.create_dataset('indptr', data=var_val.indptr)
                    hf_csr.attrs['shape'] = var_val.shape
            elif isinstance(var_val, str):
                hf_str = hf_data.create_dataset(var_name, (1,), dtype=str_dt)
                hf_str[0] = var_val
            elif isinstance(var_val, list) and var_val and isinstance(var_val[0], str):
                hf_str = hf_data.create_dataset(var_name, (len(var_val),), dtype=str_dt)
                for i_ev, var_ev in enumerate(var_val):
                    hf_str[i_ev] = var_ev
            elif var_val is not None and var_name != 'pool':
                hf_data.create_dataset(var_name, data=var_val)
        hf_data.close()

    def read_hdf5(self, *args, **kwargs):
        """This function is deprecated, use Hazard.from_hdf5."""
        LOGGER.warning("The use of Hazard.read_hdf5 is deprecated."
                       "Use Hazard.from_hdf5 instead.")
        self.__dict__ = Hazard.from_hdf5(*args, **kwargs).__dict__

    @classmethod
    def from_hdf5(cls, file_name):
        """Read hazard in hdf5 format.

        Parameters
        ----------
        file_name: str
            file name to read, with h5 format

        Returns
        -------
        haz : climada.hazard.Hazard
            Hazard object from the provided MATLAB file

        """
        LOGGER.info('Reading %s', file_name)
        haz = cls()
        hf_data = h5py.File(file_name, 'r')
        for (var_name, var_val) in haz.__dict__.items():
            if var_name != 'tag' and var_name not in hf_data.keys():
                continue
            if var_name == 'centroids':
                haz.centroids = Centroids.from_hdf5(hf_data.get(var_name))
            elif var_name == 'tag':
                haz.tag.haz_type = u_hdf5.to_string(hf_data.get('haz_type')[0])
                haz.tag.file_name = u_hdf5.to_string(hf_data.get('file_name')[0])
                haz.tag.description = u_hdf5.to_string(hf_data.get('description')[0])
            elif isinstance(var_val, np.ndarray) and var_val.ndim == 1:
                setattr(haz, var_name, np.array(hf_data.get(var_name)))
            elif isinstance(var_val, sparse.csr_matrix):
                hf_csr = hf_data.get(var_name)
                if isinstance(hf_csr, h5py.Dataset):
                    setattr(haz, var_name, sparse.csr_matrix(hf_csr))
                else:
                    setattr(haz, var_name, sparse.csr_matrix((hf_csr['data'][:],
                                                               hf_csr['indices'][:],
                                                               hf_csr['indptr'][:]),
                                                              hf_csr.attrs['shape']))
            elif isinstance(var_val, str):
                setattr(haz, var_name, u_hdf5.to_string(hf_data.get(var_name)[0]))
            elif isinstance(var_val, list):
                var_value = [x for x in map(u_hdf5.to_string, np.array(hf_data.get(var_name)).tolist())]
                setattr(haz, var_name, var_value)
            else:
                setattr(haz, var_name, hf_data.get(var_name))

        hf_data.close()
        return haz

    def _set_coords_centroids(self):
        """If centroids are raster, set lat and lon coordinates"""
        if self.centroids.meta and not self.centroids.coord.size:
            self.centroids.set_meta_to_lat_lon()

    def _events_set(self):
        """Generate set of tuples with (event_name, event_date)"""
        ev_set = set()
        for ev_name, ev_date in zip(self.event_name, self.date):
            ev_set.add((ev_name, ev_date))
        return ev_set

    def _event_plot(self, event_id, mat_var, col_name, smooth, crs_espg, axis=None,
                    figsize=(9, 13), adapt_fontsize=True, **kwargs):
        """Plot an event of the input matrix.

        Parameters
        ----------
        event_id: int or np.array(int)
            If event_id > 0, plot mat_var of
            event with id = event_id. If event_id = 0, plot maximum
            mat_var in each centroid. If event_id < 0, plot
            abs(event_id)-largest event.
        mat_var: sparse matrix
            Sparse matrix where each row is an event
        col_name: sparse matrix
            Colorbar label
        smooth: bool, optional
            smooth plot to plot.RESOLUTIONxplot.RESOLUTION
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        figsize: tuple, optional
            figure size for plt.subplots
        kwargs: optional
            arguments for pcolormesh matplotlib function

        Returns
        -------
            matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        """
        if not isinstance(event_id, np.ndarray):
            event_id = np.array([event_id])
        array_val = list()
        l_title = list()
        for ev_id in event_id:
            if ev_id > 0:
                try:
                    event_pos = np.where(self.event_id == ev_id)[0][0]
                except IndexError as err:
                    raise ValueError(f'Wrong event id: {ev_id}.') from err
                im_val = mat_var[event_pos, :].toarray().transpose()
                title = 'Event ID %s: %s' % (str(self.event_id[event_pos]),
                                             self.event_name[event_pos])
            elif ev_id < 0:
                max_inten = np.asarray(np.sum(mat_var, axis=1)).reshape(-1)
                event_pos = np.argpartition(max_inten, ev_id)[ev_id:]
                event_pos = event_pos[np.argsort(max_inten[event_pos])][0]
                im_val = mat_var[event_pos, :].toarray().transpose()
                title = '%s-largest Event. ID %s: %s' % (np.abs(ev_id),
                                                         str(self.event_id[event_pos]),
                                                         self.event_name[event_pos])
            else:
                im_val = np.max(mat_var, axis=0).toarray().transpose()
                title = '%s max intensity at each point' % self.tag.haz_type

            array_val.append(im_val)
            l_title.append(title)

        return u_plot.geo_im_from_array(array_val, self.centroids.coord, col_name,
                                        l_title, smooth=smooth, axes=axis,
                                        figsize=figsize, proj=crs_espg, adapt_fontsize=adapt_fontsize, **kwargs)

    def _centr_plot(self, centr_idx, mat_var, col_name, axis=None, **kwargs):
        """Plot a centroid of the input matrix.

        Parameters
        ----------
        centr_id: int
            If centr_id > 0, plot mat_var
            of all events at centroid with id = centr_id. If centr_id = 0,
            plot maximum mat_var of each event. If centr_id < 0,
            plot abs(centr_id)-largest centroid where highest mat_var
            are reached.
        mat_var: sparse matrix
            Sparse matrix where each column represents
            a centroid
        col_name: sparse matrix
            Colorbar label
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs: optional
            arguments for plot matplotlib function

        Returns
        -------
            matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        """
        coord = self.centroids.coord
        if centr_idx > 0:
            try:
                centr_pos = centr_idx
            except IndexError as err:
                raise ValueError(f'Wrong centroid id: {centr_idx}.') from err
            array_val = mat_var[:, centr_pos].toarray()
            title = 'Centroid %s: (%s, %s)' % (str(centr_idx),
                                               coord[centr_pos, 0],
                                               coord[centr_pos, 1])
        elif centr_idx < 0:
            max_inten = np.asarray(np.sum(mat_var, axis=0)).reshape(-1)
            centr_pos = np.argpartition(max_inten, centr_idx)[centr_idx:]
            centr_pos = centr_pos[np.argsort(max_inten[centr_pos])][0]
            array_val = mat_var[:, centr_pos].toarray()

            title = '%s-largest Centroid. %s: (%s, %s)' % \
                    (np.abs(centr_idx), str(centr_pos), coord[centr_pos, 0],
                     coord[centr_pos, 1])
        else:
            array_val = np.max(mat_var, axis=1).toarray()
            title = '%s max intensity at each event' % self.tag.haz_type

        if not axis:
            _, axis = plt.subplots(1)
        if 'color' not in kwargs:
            kwargs['color'] = 'b'
        axis.set_title(title)
        axis.set_xlabel('Event number')
        axis.set_ylabel(str(col_name))
        axis.plot(range(len(array_val)), array_val, **kwargs)
        axis.set_xlim([0, len(array_val)])
        return axis

    def _loc_return_inten(self, return_periods, inten, exc_inten):
        """Compute local exceedence intensity for given return period.

        Parameters
        ----------
        return_periods: np.array
            return periods to consider
        cen_pos: int
            centroid position

        Returns
        -------
            np.array
        """
        # sorted intensity
        sort_pos = np.argsort(inten, axis=0)[::-1, :]
        columns = np.ones(inten.shape, int)
        # pylint: disable=unsubscriptable-object  # pylint/issues/3139
        columns *= np.arange(columns.shape[1])
        inten_sort = inten[sort_pos, columns]
        # cummulative frequency at sorted intensity
        freq_sort = self.frequency[sort_pos]
        np.cumsum(freq_sort, axis=0, out=freq_sort)

        for cen_idx in range(inten.shape[1]):
            exc_inten[:, cen_idx] = self._cen_return_inten(
                inten_sort[:, cen_idx], freq_sort[:, cen_idx],
                self.intensity_thres, return_periods)

    def _check_events(self):
        """Check that all attributes but centroids contain consistent data.
        Put default date, event_name and orig if not provided. Check not
        repeated events (i.e. with same date and name)

        Raises
        ------
            ValueError
        """
        num_ev = len(self.event_id)
        num_cen = self.centroids.size
        if np.unique(self.event_id).size != num_ev:
            raise ValueError("There are events with the same identifier.")

        u_check.check_oligatories(self.__dict__, self.vars_oblig, 'Hazard.',
                                  num_ev, num_ev, num_cen)
        u_check.check_optionals(self.__dict__, self.vars_opt, 'Hazard.', num_ev)
        self.event_name = u_check.array_default(num_ev, self.event_name,
                                                'Hazard.event_name', list(self.event_id))
        self.date = u_check.array_default(num_ev, self.date, 'Hazard.date',
                                          np.ones(self.event_id.shape, dtype=int))
        self.orig = u_check.array_default(num_ev, self.orig, 'Hazard.orig',
                                          np.zeros(self.event_id.shape, dtype=bool))
        if len(self._events_set()) != num_ev:
            raise ValueError("There are events with same date and name.")

    @staticmethod
    def _cen_return_inten(inten, freq, inten_th, return_periods):
        """From ordered intensity and cummulative frequency at centroid, get
        exceedance intensity at input return periods.

        Parameters
        ----------
        inten: np.array
            sorted intensity at centroid
        freq: np.array
            cummulative frequency at centroid
        inten_th: float
            intensity threshold
        return_periods: np.array
            return periods

        Returns
        -------
            np.array
        """
        inten_th = np.asarray(inten > inten_th).squeeze()
        inten_cen = inten[inten_th]
        freq_cen = freq[inten_th]
        if not inten_cen.size:
            return np.zeros((return_periods.size,))
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                pol_coef = np.polyfit(np.log(freq_cen), inten_cen, deg=1)
        except ValueError:
            pol_coef = np.polyfit(np.log(freq_cen), inten_cen, deg=0)
        inten_fit = np.polyval(pol_coef, np.log(1 / return_periods))
        wrong_inten = (return_periods > np.max(1 / freq_cen)) & np.isnan(inten_fit)
        inten_fit[wrong_inten] = 0.

        return inten_fit

    def _read_att_mat(self, data, file_name, var_names):
        """Read MATLAB hazard's attributes."""
        self.frequency = np.squeeze(data[var_names['var_name']['freq']])
        self.orig = np.squeeze(data[var_names['var_name']['orig']]).astype(bool)
        self.event_id = np.squeeze(
            data[var_names['var_name']['even_id']].astype(int, copy=False))
        try:
            self.units = u_hdf5.get_string(data[var_names['var_name']['unit']])
        except KeyError:
            pass

        n_cen = self.centroids.size
        n_event = len(self.event_id)
        try:
            self.intensity = u_hdf5.get_sparse_csr_mat(
                data[var_names['var_name']['inten']], (n_event, n_cen))
        except ValueError as err:
            raise ValueError('Size missmatch in intensity matrix.') from err
        try:
            self.fraction = u_hdf5.get_sparse_csr_mat(
                data[var_names['var_name']['frac']], (n_event, n_cen))
        except ValueError as err:
            raise ValueError('Size missmatch in fraction matrix.') from err
        except KeyError:
            self.fraction = sparse.csr_matrix(np.ones(self.intensity.shape, dtype=float))
        # Event names: set as event_id if no provided
        try:
            self.event_name = u_hdf5.get_list_str_from_ref(
                file_name, data[var_names['var_name']['ev_name']])
        except KeyError:
            self.event_name = list(self.event_id)
        try:
            comment = u_hdf5.get_string(data[var_names['var_name']['comment']])
            self.tag.description += ' ' + comment
        except KeyError:
            pass

        try:
            datenum = data[var_names['var_name']['datenum']].squeeze()
            self.date = np.array([
                (dt.datetime.fromordinal(int(date))
                 + dt.timedelta(days=date % 1)
                 - dt.timedelta(days=366)).toordinal()
                for date in datenum])
        except KeyError:
            pass

    def _read_att_excel(self, file_name, var_names):
        """Read Excel hazard's attributes."""
        dfr = pd.read_excel(file_name, var_names['sheet_name']['freq'])

        num_events = dfr.shape[0]
        self.frequency = dfr[var_names['col_name']['freq']].values
        self.orig = dfr[var_names['col_name']['orig']].values.astype(bool)
        self.event_id = dfr[var_names['col_name']['even_id']].values. \
            astype(int, copy=False)
        self.date = dfr[var_names['col_name']['even_dt']].values. \
            astype(int, copy=False)
        self.event_name = dfr[var_names['col_name']['even_name']].values.tolist()

        dfr = pd.read_excel(file_name, var_names['sheet_name']['inten'])
        # number of events (ignore centroid_ID column)
        # check the number of events is the same as the one in the frequency
        if dfr.shape[1] - 1 is not num_events:
            raise ValueError('Hazard intensity is given for a number of events '
                             'different from the number of defined in its frequency: '
                             f'{dfr.shape[1] - 1} != {num_events}')
        # check number of centroids is the same as retrieved before
        if dfr.shape[0] is not self.centroids.size:
            raise ValueError('Hazard intensity is given for a number of centroids '
                             'different from the number of centroids defined: '
                             f'{dfr.shape[0]} != {self.centroids.size}')

        self.intensity = sparse.csr_matrix(dfr.values[:, 1:num_events + 1].transpose())
        self.fraction = sparse.csr_matrix(np.ones(self.intensity.shape, dtype=float))

    def append(self, *others):
        """Append the events and centroids to this hazard object.

        All of the given hazards must be of the same type and use the same units as self. The
        centroids of all hazards must have the same CRS.

        The following kinds of object attributes are processed:

        - All centroids are combined together using `Centroids.union`.

        - Lists, 1-dimensional arrays (NumPy) and sparse CSR matrices (SciPy) are concatenated.
        Sparse matrices are concatenated along the first (vertical) axis.

        - All `tag` attributes are appended to `self.tag`.

        For any other type of attribute: A ValueError is raised if an attribute of that name is
        not defined in all of the non-empty hazards at least. However, there is no check that the
        attribute value is identical among the given hazard objects. The initial attribute value of
        `self` will not be modified.

        Note: Each of the hazard's `centroids` attributes might be modified in place in the sense
        that missing properties are added, but existing ones are not overwritten. In case of raster
        centroids, conversion to point centroids is applied so that raster information (meta) is
        lost. For more information, see `Centroids.union`.

        Parameters
        ----------
        others : one or more climada.hazard.Hazard objects
            Hazard instances to append to self

        Raises
        ------
        TypeError, ValueError

        See Also
        --------
        Hazard.concat : concatenate 2 or more hazards
        Centroids.union : combine centroids
        """
        if len(others) == 0:
            return
        haz_list = [self] + list(others)
        haz_list_nonempty = [haz for haz in haz_list if haz.size > 0]

        for haz in haz_list:
            haz._check_events()

        # check type, unit, and attribute consistency among hazards
        haz_types = {haz.tag.haz_type for haz in haz_list if haz.tag.haz_type != ''}
        if len(haz_types) > 1:
            raise ValueError(f"The given hazards are of different types: {haz_types}. "
                             "The hazards are incompatible and cannot be concatenated.")
        self.tag.haz_type = haz_types.pop()

        haz_classes = {type(haz) for haz in haz_list}
        if len(haz_classes) > 1:
            raise TypeError(f"The given hazards are of different classes: {haz_classes}. "
                            "The hazards are incompatible and cannot be concatenated.")

        units = {haz.units for haz in haz_list if haz.units != ''}
        if len(units) > 1:
            raise ValueError(f"The given hazards use different units: {units}. "
                             "The hazards are incompatible and cannot be concatenated.")
        elif len(units) == 0:
            units = {''}
        self.units = units.pop()

        attributes = sorted(set.union(*[set(vars(haz).keys()) for haz in haz_list]))
        for attr_name in attributes:
            if not all(hasattr(haz, attr_name) for haz in haz_list_nonempty):
                raise ValueError(f"Attribute {attr_name} is not shared by all hazards. "
                                 "The hazards are incompatible and cannot be concatenated.")

        # append all tags (to keep track of input files and descriptions)
        for haz in haz_list:
            if haz.tag is not self.tag:
                self.tag.append(haz.tag)

        # map individual centroids objects to union
        centroids = Centroids.union(*[haz.centroids for haz in haz_list])
        hazcent_in_cent_idx_list = [
            u_coord.assign_coordinates(haz.centroids.coord, centroids.coord, threshold=0)
            for haz in haz_list_nonempty
        ]

        # concatenate array and list attributes of non-empty hazards
        for attr_name in attributes:
            attr_val_list = [getattr(haz, attr_name) for haz in haz_list_nonempty]
            if isinstance(attr_val_list[0], sparse.csr.csr_matrix):
                # map sparse matrix onto centroids
                setattr(self, attr_name, sparse.vstack([
                    sparse.csr_matrix(
                        (matrix.data, cent_idx[matrix.indices], matrix.indptr),
                        shape=(matrix.shape[0], centroids.size)
                    )
                    for matrix, cent_idx in zip(attr_val_list, hazcent_in_cent_idx_list)
                ], format='csr'))
            elif isinstance(attr_val_list[0], np.ndarray) and attr_val_list[0].ndim == 1:
                setattr(self, attr_name, np.hstack(attr_val_list))
            elif isinstance(attr_val_list[0], list):
                setattr(self, attr_name, sum(attr_val_list, []))

        self.centroids = centroids
        self.sanitize_event_ids()

    @classmethod
    def concat(cls, haz_list):
        """
        Concatenate events of several hazards of same type.

        This function creates a new hazard of the same class as the first hazard in the given list
        and then applies the `append` method. Please refer to the docs of `Hazard.append` for
        caveats and limitations of the concatenation procedure.

        For centroids, tags, lists, arrays and sparse matrices, the remarks in `Hazard.append`
        apply. All other attributes are copied from the first object in `haz_list`.

        Note that `Hazard.concat` can be used to concatenate hazards of a subclass. The result's
        type will be the subclass. However, calling `concat([])` (with an empty list) is equivalent
        to instantiation without init parameters. So, `Hazard.concat([])` is equivalent to
        `Hazard()`. If `HazardB` is a subclass of `Hazard`, then `HazardB.concat([])` is equivalent
        to `HazardB()` (unless `HazardB` overrides the `concat` method).

        Parameters
        ----------
        haz_list : list of climada.hazard.Hazard objects
            Hazard instances of the same hazard type (subclass).

        Returns
        -------
        haz_concat : instance of climada.hazard.Hazard
            This will be of the same type (subclass) as all the hazards in `haz_list`.

        See Also
        --------
        Hazard.append : append hazards to a hazard in place
        Centroids.union : combine centroids
        """
        if len(haz_list) == 0:
            return cls()
        haz_concat = haz_list[0].__class__()
        haz_concat.tag.haz_type = haz_list[0].tag.haz_type
        for attr_name, attr_val in vars(haz_list[0]).items():
            # to save memory, only copy simple attributes like
            # "units" that are not explicitly handled by Hazard.append
            if not (isinstance(attr_val, (list, np.ndarray, sparse.csr.csr_matrix))
                    or attr_name in ["tag", "centroids"]):
                setattr(haz_concat, attr_name, copy.deepcopy(attr_val))
        haz_concat.append(*haz_list)
        return haz_concat

    def change_centroids(self, centroids, threshold=NEAREST_NEIGHBOR_THRESHOLD):
        """
        Assign (new) centroids to hazard.

        Centoids of the hazard not in centroids are mapped onto the nearest
        point. Fails if a point is further than threshold from the closest
        centroid.

        The centroids must have the same CRS as self.centroids.

        Parameters
        ----------
        haz: Hazard
            Hazard instance
        centroids: Centroids
            Centroids instance on which to map the hazard.
        threshold: int or float
            Threshold (in km) for mapping haz.centroids not in centroids.
            Argument is passed to climada.util.coordinates.assign_coordinates.
            Default: 100 (km)

        Returns
        -------
        haz_new_cent: Hazard
            Hazard projected onto centroids

        Raises
        ------
        ValueError


        See Also
        --------
        util.coordinates.assign_coordinates: algorithm to match centroids.

        """
        # define empty hazard
        haz_new_cent = copy.deepcopy(self)
        haz_new_cent.centroids = centroids

        # indices for mapping matrices onto common centroids
        if centroids.meta:
            new_cent_idx = u_coord.assign_grid_points(
                self.centroids.lon, self.centroids.lat,
                centroids.meta['width'], centroids.meta['height'],
                centroids.meta['transform'])
            if -1 in new_cent_idx:
                raise ValueError("At least one hazard centroid is out of"
                                 "the raster defined by centroids.meta."
                                 " Please choose a larger raster.")
        else:
            new_cent_idx = u_coord.assign_coordinates(
                self.centroids.coord, centroids.coord, threshold=threshold
            )
            if -1 in new_cent_idx:
                raise ValueError("At least one hazard centroid is at a larger "
                                 f"distance than the given threshold {threshold} "
                                 "from the given centroids. Please choose a "
                                 "larger threshold or enlarge the centroids")

        if np.unique(new_cent_idx).size < new_cent_idx.size:
            raise ValueError("At least two hazard centroids are mapped to the same "
                             "centroids. Please make sure that the given centroids "
                             "cover the same area like the original centroids and "
                             "are not of lower resolution.")

        # re-assign attributes intensity and fraction
        for attr_name in ["intensity", "fraction"]:
            matrix = getattr(self, attr_name)
            setattr(haz_new_cent, attr_name,
                    sparse.csr_matrix(
                        (matrix.data, new_cent_idx[matrix.indices], matrix.indptr),
                        shape=(matrix.shape[0], centroids.size)
                    ))

        return haz_new_cent



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Unit Tests for LitPop class.
"""

import numpy as np
import unittest
from rasterio.crs import CRS
from rasterio import Affine
from climada.entity.exposures.litpop import litpop as lp


def data_arrays_demo(number_of_arrays=2):
    """init demo data arrays (2d) for LitPop core calculations"""
    data_arrays = list()
    if number_of_arrays > 0:
        data_arrays.append(np.array([[0,1,2], [3,4,5]]))
        # array([[0, 1, 2],
        #       [3, 4, 5]])
    if number_of_arrays > 1:
        data_arrays.append(np.array([[10,10,10], [1,1,1]]))
        # array([[10, 10, 10],
        #       [1, 1, 1]])
    if number_of_arrays > 2:
        data_arrays.append(np.array([[0,1,10], [0,1,10]]))
        # array([[0, 1, 10],
        #       [0, 1, 10]])
    if number_of_arrays > 3:
        data_arrays.append([[0,1,10,100], [0,1,10,100]])
        # [[0, 1, 10, 100],
        #  [0, 1, 10, 100]]
    return data_arrays

def data_arrays_resampling_demo():
    """init demo data arrays (2d) and meta data for resampling"""
    data_arrays = list()
    # demo pop:
    data_arrays.append(np.array([[0,1,2], [3,4,5]], dtype='float32'))
    data_arrays.append(np.array([[0,1,2], [3,4,5]], dtype='float32'))
        # array([[0, 1, 2],
        #       [3, 4, 5]])
    # demo nightlight:
    data_arrays.append(np.array([[2,10,0, 0, 0, 0], [10,2,10, 0, 0, 0],
                                 [0,0,0, 0, 1, 1], [1,0,0, 0, 1, 1]],
                       dtype='float32'))
        # array([[ 2., 10.,  0.,  0.,  0.,  0.],
        #        [10.,  2., 10.,  0.,  0.,  0.],
        #        [ 0.,  0.,  0.,  0.,  1.,  1.],
        #        [ 1.,  0.,  0.,  0.,  1.,  1.]], dtype=float32)]

    meta_list = [{'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': -3.4028230607370965e+38,
                  'width': 3,
                  'height': 2,
                  'count': 1,
                  'crs': CRS.from_epsg(4326),
                  #'crs': CRS.from_epsg(4326),
                  'transform': Affine(1, 0.0, -10,
                         0.0, -1, 40),
                  },
                 {'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': -3.4028230607370965e+38,
                  'width': 3,
                  'height': 2,
                  'count': 1,
                  'crs': CRS.from_epsg(4326),
                  #'crs': CRS.from_epsg(4326),
                  'transform': Affine(1, 0.0, -10,
                         0.0, -1, 41), # shifted by 1 degree latitude to the north
                  },
                 {'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': None,
                  'width': 6,
                  'height': 4,
                  'count': 1,
                  'crs': CRS.from_epsg(4326),
                  # 'crs': CRS.from_epsg(32662),
                  'transform': Affine(.5, 0.0, -10,
                         0.0, -.5, 40), # higher resolution
                  }]
    return data_arrays, meta_list


class TestLitPop(unittest.TestCase):
    """Test LitPop Class methods and functions"""

    def test_reproject_input_data_downsample(self):
        """test function reproject_input_data downsampling lit to pop grid
        (default resampling for LitPop)"""
        data_in, meta_list = data_arrays_resampling_demo()
        #
        data_out, meta_out = lp.reproject_input_data(data_in, meta_list,
                        i_align=0,
                        target_res_arcsec=None,
                        global_origins=(-180, 90)
                        )
        # test reference data unchanged:
        np.testing.assert_array_equal(data_in[0], data_out[0])
        # test northward shift:
        np.testing.assert_array_equal(data_in[1][1,:], data_out[1][0,:])
        # test reprojected nl data:
        reference_array = np.array([[5.020408  , 2.267857  , 0.12244898],
                                    [1.1224489 , 0.6785714 , 0.7346939 ]], dtype='float32')
        np.testing.assert_array_almost_equal_nulp(reference_array, data_out[2])

    def test_reproject_input_data_downsample_conserve_sum(self):
        """test function reproject_input_data downsampling with conservation of sum"""
        data_in, meta_list = data_arrays_resampling_demo()
        #
        data_out, meta_out = lp.reproject_input_data(data_in, meta_list,
                        i_align=0,
                        target_res_arcsec=None,
                        global_origins=(-180, 90),
                        conserve='sum')
        # test reference data unchanged:
        np.testing.assert_array_equal(data_in[0], data_out[0])
        # test conserve sum:
        for i, _ in enumerate(data_in):
            self.assertAlmostEqual(data_in[i].sum(), data_out[i].sum())

    def test_reproject_input_data_downsample_conserve_mean(self):
        """test function reproject_input_data downsampling with conservation of sum"""
        data_in, meta_list = data_arrays_resampling_demo()
        #
        data_out, meta_out = lp.reproject_input_data(data_in, meta_list,
                        i_align=1,
                        target_res_arcsec=None,
                        global_origins=(-180, 90),
                        conserve='mean')
        # test reference data unchanged:
        np.testing.assert_array_equal(data_in[1], data_out[1])
        # test conserve sum:
        for i, _ in enumerate(data_in):
            self.assertAlmostEqual(data_in[i].mean(), data_out[i].mean(), places=5)

    def test_reproject_input_data_upsample(self):
        """test function reproject_input_data with upsampling
        (usually not required for LitPop)"""
        data_in, meta_list = data_arrays_resampling_demo()
        #
        data_out, meta_out = lp.reproject_input_data(data_in, meta_list,
                        i_align=2, # high res data as reference
                        target_res_arcsec=None,
                        global_origins=(-180, 90)
                        )
        # test reference data unchanged:
        np.testing.assert_array_equal(data_in[2], data_out[2])
        # test northward shift:
        np.testing.assert_array_equal(data_out[0][2,:], data_out[1][0,:])
        np.testing.assert_array_equal(data_out[0][3,:], data_out[1][1,:])
        # test reprojected nl data:
        reference_array = np.array([[0.  , 0.25, 0.75, 1.25, 1.75, 2.  ],
                                    [0.75, 1.  , 1.5 , 2.  , 2.5 , 2.75],
                                    [2.25, 2.5 , 3.  , 3.5 , 4.  , 4.25],
                                    [3.  , 3.25, 3.75, 4.25, 4.75, 5.  ]], dtype='float32')
        np.testing.assert_array_equal(reference_array, data_out[0])

    def test_reproject_input_data_odd_downsample(self):
        """test function reproject_input_data with odd downsampling"""
        data_in, meta_list = data_arrays_resampling_demo()
        #
        data_out, meta_out = \
            lp.reproject_input_data(data_in, meta_list,
                                   i_align=0, # high res data as reference
                                   target_res_arcsec=6120, # 1.7 degree
                                   global_origins=(-180, 90),
                                   )
        self.assertEqual(1.7, meta_out['transform'][0]) # check resolution
        reference_array = np.array([[0.425    , 1.7631578],
                                    [3.425    , 4.763158 ]], dtype='float32')
        np.testing.assert_array_equal(reference_array, data_out[0])

    def test_gridpoints_core_calc_input_errors(self):
        """test for ValueErrors and TypeErrors due to wrong input to function
        gridpoints_core_calc"""
        data = data_arrays_demo(2)
        # negative offset:
        with self.assertRaises(ValueError):
            lp.gridpoints_core_calc(data, offsets=[2,-1])
        # negative exponents:
        with self.assertRaises(ValueError):
            lp.gridpoints_core_calc(data, exponents=[2,-1])

        # different shapes:
        with self.assertRaises(ValueError):
            lp.gridpoints_core_calc(data_arrays_demo(4))

        # wrong format:
        with self.assertRaises(TypeError):
            lp.gridpoints_core_calc(data, exponents=['a', 'b'])
        data.append('hello i am a string')
        with self.assertRaises(ValueError):
            lp.gridpoints_core_calc(data)
        with self.assertRaises(TypeError):
            lp.gridpoints_core_calc(777)

    def test_gridpoints_core_calc_default_1(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with default exponents and offsets - 1 array"""
        data_arrays = data_arrays_demo(1) # get list with 1 demo array
        result_array = lp.gridpoints_core_calc(data_arrays)
        results_check = data_arrays[0]

        self.assertEqual(result_array.shape, results_check.shape)
        self.assertEqual(result_array[1,1], results_check[1,1])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

    def test_gridpoints_core_calc_default_2(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with default exponents and offsets- 2 arrays"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        result_array = lp.gridpoints_core_calc(data_arrays)
        results_check = data_arrays[0] * data_arrays[1]

        self.assertEqual(result_array[0,0], results_check[0,0])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)


    def test_gridpoints_core_calc_default_3(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with default exponents and offsets- 3 arrays"""
        data_arrays = data_arrays_demo(3)
        result_array = lp.gridpoints_core_calc(data_arrays)
        results_check = data_arrays[0] * data_arrays[1] * data_arrays[2]

        self.assertEqual(result_array.shape, results_check.shape)
        self.assertEqual(result_array[1,1], results_check[1,1])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)
        #self.assertEqual(result_array, data_arrays[0] * data_arrays[1])

    def test_gridpoints_core_calc_exp(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with changed exponents"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        exp = [2, 1]
        result_array = lp.gridpoints_core_calc(data_arrays, exponents=exp)
        results_check = data_arrays[0] * data_arrays[0] * data_arrays[1]

        self.assertEqual(result_array.shape, results_check.shape)
        self.assertEqual(result_array[0,2], results_check[0,2])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

        exp = [2, .1]
        result_array = lp.gridpoints_core_calc(data_arrays, exponents=exp)
        results_check = data_arrays[0] * data_arrays[0] * (data_arrays[1] ** .1)
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

    def test_gridpoints_core_calc_offsets(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with changed offsets"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        offsets = [1, 10]
        result_array = lp.gridpoints_core_calc(data_arrays, offsets=offsets)
        results_check = (data_arrays[0]+1) * (10 + data_arrays[1])

        self.assertEqual(result_array.shape, results_check.shape)
        self.assertEqual(result_array[0,2], results_check[0,2])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

    def test_gridpoints_core_calc_offsets_exp(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with changed offsets and exponents"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        offsets = [0, 10]
        exp = [2, 1]
        result_array = lp.gridpoints_core_calc(data_arrays, offsets=offsets,
                                              exponents=exp)
        results_check = (data_arrays[0]) * (data_arrays[0]) * (10+data_arrays[1])
        results_check2 = np.array([[0, 20, 80],[99, 176, 275]])

        self.assertEqual(result_array.shape, results_check.shape)
        self.assertEqual(result_array[0,2], results_check[0,2])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)
        self.assertEqual(result_array[1,2], results_check2[1,2])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check2)

    def test_gridpoints_core_calc_rescale(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with rescaling (default exponents and offsets)"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        result_array = lp.gridpoints_core_calc(data_arrays, total_val_rescale=2.5)
        results_check = (data_arrays[0]*data_arrays[1]) * 2.5/np.sum(data_arrays[0]*data_arrays[1])

        self.assertAlmostEqual(result_array.sum(), 2.5)
        self.assertEqual(result_array[0,1], results_check[0,1])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

    def test_gridpoints_core_calc_offsets_exp_rescale(self):
        """test function gridpoints_core_calc, i.e. core data combination
        on grid point level with changed offsets and exponents and rescaling"""
        data_arrays = data_arrays_demo(2) # get list with 2 demo arrays
        offsets = [0.2, 3]
        exp = [.5, 1.7]
        tot = -7
        result_array = lp.gridpoints_core_calc(data_arrays, offsets=offsets,
                                              exponents=exp, total_val_rescale=tot)
        results_check = np.array(data_arrays[0]+.2, dtype=float)**exp[0] * \
            (np.array(data_arrays[1]+3., dtype=float)**exp[1])
        results_check = results_check * tot / results_check.sum()
        self.assertEqual(result_array.shape, results_check.shape)
        self.assertAlmostEqual(result_array.sum(), tot)
        self.assertEqual(result_array[1,2], results_check[1,2])
        np.testing.assert_array_almost_equal_nulp(result_array, results_check)

    def test_grp_read_pass(self):
        """test _grp_read() to pass and return either dict with admin1 values or None"""
        result = lp._grp_read('JPN')
        if result is not None:
            self.assertIsInstance(result, dict)
            self.assertIn('Fukuoka', result.keys())
            self.assertIsInstance(result['Saga'], float)

    def test_fail_get_total_value_per_country_pop(self):
        "test _get_total_value_per_country fails for pop"
        with self.assertRaises(NotImplementedError):
            lp._get_total_value_per_country('XXX', 'pop', None)

    def test_get_total_value_per_country_none(self):
        "test _get_total_value_per_country pass with None"
        value = lp._get_total_value_per_country('XXX', 'none', None)
        self.assertEqual(value, None)

    def test_get_total_value_per_country_norm(self):
        "test _get_total_value_per_country pass with 1"
        value = lp._get_total_value_per_country('XXX', 'norm', None)
        self.assertEqual(value, 1)

    def test_get_total_value_per_country_gdp(self):
        "test _get_total_value_per_country get number for gdp"
        gdp_togo = lp._get_total_value_per_country('TGO', 'gdp', 2010)
        gdp_switzerland = lp._get_total_value_per_country('CHE', 'gdp', 2222)
        value_switzerland = lp._get_total_value_per_country('CHE', 'income_group', 2222)
        self.assertIsInstance(gdp_togo, float)
        # value for income_group = gdp * income group:
        self.assertEqual(value_switzerland, 5*gdp_switzerland)

    def test_get_total_value_per_country_pc(self):
        "test _get_total_value_per_country get number for pc of Poland"
        value = lp._get_total_value_per_country('POL', 'pc', 2015)
        self.assertIsInstance(value, float)

    def test_get_total_value_per_country_nfw(self):
        "test _get_total_value_per_country get number for pc of Poland"
        value = lp._get_total_value_per_country('POL', 'nfw', 2015)
        self.assertIsInstance(value, float)

    def test_get_value_unit_pass(self):
        """test get_value_unit pass"""
        self.assertEqual(lp.get_value_unit('pop'), 'people')
        self.assertEqual(lp.get_value_unit('gdp'), 'USD')
        self.assertEqual(lp.get_value_unit('pc'), 'USD')
        self.assertEqual(lp.get_value_unit('nfw'), 'USD')
        self.assertEqual(lp.get_value_unit('none'), '')

if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestLitPop)
    # TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestUncertainty))
    unittest.TextTestRunner(verbosity=2).run(TESTS)


"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Tests on LitPop exposures.
"""
import unittest
import numpy as np
from shapely.geometry import Polygon

from climada.entity.exposures.litpop import litpop as lp
from climada.entity.exposures.litpop import gpw_population
from climada.util.finance import world_bank_wealth_account, gdp, income_group
import climada.util.coordinates as u_coord
from climada.util.constants import SYSTEM_DIR
from climada import CONFIG


class TestLitPopExposure(unittest.TestCase):
    """Test LitPop exposure data model:"""

    def test_netherlands150_pass(self):
        """Test from_countries for Netherlands at 150 arcsec, first shape is empty"""
        ent = lp.LitPop.from_countries('Netherlands', res_arcsec=150, reference_year=2016)
        self.assertEqual(ent.gdf.shape[0], 2829)

    def test_BLM150_pass(self):
        """Test from_countries for BLM at 150 arcsec, 2 data points"""
        ent = lp.LitPop.from_countries('BLM', res_arcsec=150, reference_year=2016)
        self.assertEqual(ent.gdf.shape[0], 2)

    def test_Monaco150_pass(self):
        """Test from_countries for Moncao at 150 arcsec, 1 data point"""
        ent = lp.LitPop.from_countries('Monaco', res_arcsec=150, reference_year=2016)
        self.assertEqual(ent.gdf.shape[0], 1)

    def test_switzerland300_pass(self):
        """Create LitPop entity for Switzerland on 300 arcsec:"""
        country_name = ['CHE']
        resolution = 300
        fin_mode = 'income_group'
        with self.assertLogs('climada.entity.exposures.litpop', level='INFO') as cm:
            ent = lp.LitPop.from_countries(country_name, res_arcsec=resolution, fin_mode=fin_mode,
                                           reference_year=2016)

        self.assertIn('LitPop: Init Exposure for country: CHE', cm.output[0])
        self.assertEqual(ent.gdf.region_id.min(), 756)
        self.assertEqual(ent.gdf.region_id.max(), 756)
        # confirm that the total value is equal to GDP * (income_group+1):
        self.assertAlmostEqual(ent.gdf.value.sum()/gdp('CHE', 2016)[1],
                               (income_group('CHE', 2016)[1] + 1))
        self.assertIn("LitPop Exposure for ['CHE'] at 300 as, year: 2016", ent.tag.description)
        self.assertIn('income_group', ent.tag.description)
        self.assertIn('1, 1', ent.tag.description)
        self.assertTrue(u_coord.equal_crs(ent.crs, 'epsg:4326'))
        self.assertEqual(ent.meta['width'], 54)
        self.assertEqual(ent.meta['height'], 23)
        self.assertTrue(u_coord.equal_crs(ent.meta['crs'], 'epsg:4326'))
        self.assertAlmostEqual(ent.meta['transform'][0], 0.08333333333333333)
        self.assertAlmostEqual(ent.meta['transform'][1], 0)
        self.assertAlmostEqual(ent.meta['transform'][2], 5.9166666666666)
        self.assertAlmostEqual(ent.meta['transform'][3], 0)
        self.assertAlmostEqual(ent.meta['transform'][4], -0.08333333333333333)
        self.assertAlmostEqual(ent.meta['transform'][5], 47.75)

    def test_switzerland30normPop_pass(self):
        """Create LitPop entity for Switzerland on 30 arcsec:"""
        country_name = ['CHE']
        resolution = 30
        exp = [0, 1]
        fin_mode = 'norm'
        with self.assertLogs('climada.entity.exposures.litpop', level='INFO') as cm:
            ent = lp.LitPop.from_countries(country_name, res_arcsec=resolution, exponents=exp,
                                           fin_mode=fin_mode, reference_year=2015)
        # print(cm)
        self.assertIn('LitPop: Init Exposure for country: CHE', cm.output[0])
        self.assertEqual(ent.gdf.region_id.min(), 756)
        self.assertEqual(ent.gdf.region_id.max(), 756)
        self.assertEqual(ent.gdf.value.sum(), 1.0)
        self.assertEqual(ent.ref_year, 2015)

    def test_suriname30_nfw_pass(self):
        """Create LitPop entity for Suriname for non-finanical wealth in 2016:"""
        country_name = ['SUR']
        fin_mode = 'nfw'
        ent = lp.LitPop.from_countries(country_name, reference_year=2016, fin_mode=fin_mode)

        self.assertEqual(ent.gdf.region_id.min(), 740)
        self.assertEqual(ent.gdf.region_id.max(), 740)
        self.assertEqual(ent.ref_year, 2016)

    def test_switzerland300_admin1_pc2016_pass(self):
        """Create LitPop entity for Switzerland 2016 with admin1 and produced capital:"""
        country_name = ['CHE']
        fin_mode = 'pc'
        resolution = 300
        ref_year = 2016
        adm1 = True
        comparison_total_val = world_bank_wealth_account(country_name[0], ref_year, no_land=1)[1]
        ent = lp.LitPop.from_countries(country_name, res_arcsec=resolution,
                                       reference_year=ref_year, fin_mode=fin_mode,
                                       admin1_calc=adm1)

        self.assertAlmostEqual(np.around(ent.gdf.value.sum()*1e-9, 0),
                         np.around(comparison_total_val*1e-9, 0), places=0)
        self.assertEqual(ent.value_unit, 'USD')

    def test_from_shape_zurich_pass(self):
        """test initiating LitPop for custom shape (square around Zurich City)
        Distributing an imaginary total value of 1000 USD"""
        bounds = (8.41, 47.2, 8.70, 47.45) # (min_lon, max_lon, min_lat, max_lat)
        total_value=1000
        shape = Polygon([
            (bounds[0], bounds[3]),
            (bounds[2], bounds[3]),
            (bounds[2], bounds[1]),
            (bounds[0], bounds[1])
            ])
        ent = lp.LitPop.from_shape(shape, total_value, res_arcsec=30, reference_year=2016)
        self.assertEqual(ent.gdf.value.sum(), 1000.0)
        self.assertEqual(ent.gdf.value.min(), 0.0)
        self.assertEqual(ent.gdf.region_id.min(), 756)
        self.assertEqual(ent.gdf.region_id.max(), 756)
        self.assertAlmostEqual(ent.gdf.latitude.min(), 47.20416666666661)
        # index and coord. of largest value:
        self.assertEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].index[0], 482)
        self.assertAlmostEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].latitude.values[0], 47.34583333333325)
        self.assertAlmostEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].longitude.values[0], 8.529166666666658)

    def test_from_shape_and_countries_zurich_pass(self):
        """test initiating LitPop for custom shape (square around Zurich City)
        with from_shape_and_countries()"""
        bounds = (8.41, 47.2, 8.70, 47.45) # (min_lon, max_lon, min_lat, max_lat)
        shape = Polygon([
            (bounds[0], bounds[3]),
            (bounds[2], bounds[3]),
            (bounds[2], bounds[1]),
            (bounds[0], bounds[1])
            ])
        ent = lp.LitPop.from_shape_and_countries(
            shape, 'Switzerland', res_arcsec=30, reference_year=2016)
        self.assertEqual(ent.gdf.value.min(), 0.0)
        self.assertEqual(ent.gdf.region_id.min(), 756)
        self.assertEqual(ent.gdf.region_id.max(), 756)
        self.assertAlmostEqual(ent.gdf.latitude.min(), 47.20416666666661)
        # coord of largest value:
        self.assertEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].index[0], 434)
        self.assertAlmostEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].latitude.values[0], 47.34583333333325)
        self.assertAlmostEqual(ent.gdf.loc[ent.gdf.value == ent.gdf.value.max()].longitude.values[0], 8.529166666666658)

    def test_Liechtenstein_15_lit_pass(self):
        """Create Nightlights entity for Liechtenstein 2016:"""
        country_name = 'Liechtenstein'
        ref_year = 2016
        ent = lp.LitPop.from_nightlight_intensity(country_name, reference_year=ref_year)

        self.assertEqual(ent.gdf.value.sum(), 36469.0)
        self.assertEqual(ent.gdf.region_id[1], 438)
        self.assertEqual(ent.value_unit, '')
        self.assertAlmostEqual(ent.gdf.latitude.max(), 47.260416666666664)
        self.assertAlmostEqual(ent.meta['transform'][4], -15/3600)

    def test_Liechtenstein_30_pop_pass(self):
        """Create population count entity for Liechtenstein 2015:"""
        country_name = 'Liechtenstein'
        ref_year = 2015
        ent = lp.LitPop.from_population(country_name, reference_year=ref_year)

        self.assertEqual(ent.gdf.value.sum(), 30068.970703125)
        self.assertEqual(ent.gdf.region_id[1], 438)
        self.assertEqual(ent.value_unit, 'people')
        self.assertAlmostEqual(ent.gdf.latitude.max(), 47.2541666666666)
        self.assertAlmostEqual(ent.meta['transform'][0], 30/3600)

class TestAdmin1(unittest.TestCase):
    """Test the admin1 functionalities within the LitPop module"""

    def test_from_countries_calc_admin1_pass(self):
        """test method from_countries with admin1_calc=True for Switzerland"""
        country = "Switzerland"
        resolution = 90
        fin_mode = 'gdp'

        ent = lp.LitPop.from_countries(country, res_arcsec=resolution, fin_mode=fin_mode,
                                       reference_year=2016, admin1_calc=True)
        ent_adm0 = lp.LitPop.from_countries(country, res_arcsec=resolution, fin_mode=fin_mode,
                                            reference_year=2016, admin1_calc=False)
        # shape must be same as with admin1_calc = False, otherwise there
        # is a problem with handling of the admin1 shapes:
        self.assertEqual(ent.gdf.shape[0], 7800)
        self.assertEqual(ent.gdf.shape[0], ent_adm0.gdf.shape[0])

    def test_calc_admin1(self):
        """test function _calc_admin1_one_country for Switzerland."""
        resolution = 300
        country = 'CHE'
        ent = lp._calc_admin1_one_country(country, resolution, (2,1), 'pc', None,
                 2016, lp.GPW_VERSION, SYSTEM_DIR)

        self.assertEqual(ent.gdf.shape[0], 699)
        self.assertEqual(ent.gdf.region_id[88], 756)
        self.assertAlmostEqual(ent.gdf.latitude.max(), 47.708333333333336)
        # shape must be same as with admin1_calc = False, otherwise there
        # is a problem with handling of the admin1 shapes:
        ent_adm0 = lp.LitPop.from_countries(country, res_arcsec=resolution, fin_mode='pc',
                                            reference_year=2016, admin1_calc=False)
        self.assertEqual(ent.gdf.shape[0], ent_adm0.gdf.shape[0])

    def test_brandenburg(self):
        """test functions from_shape_and_countries and from_shape
        for admin1 shape of Brandenburg"""
        reslution_arcsec = 120
        country = 'DEU'
        state_name = 'Brandenburg'
        # get the shape of Brandenburg:
        admin1_info, admin1_shapes = u_coord.get_admin1_info(country)
        admin1_info = admin1_info[country]
        admin1_shapes = admin1_shapes[country]
        admin1_names = [record['name'] for record in admin1_info]
        print(admin1_names)
        for idx, name in enumerate(admin1_names):
            if admin1_names[idx]==state_name:
                break

        # init LitPop for Brandenburg
        exp_bra2 = lp.LitPop.from_shape_and_countries(
            admin1_shapes[idx], country, res_arcsec=reslution_arcsec, reference_year=2016)
        exp_bra = lp.LitPop.from_shape(
            admin1_shapes[idx], 1000, res_arcsec=reslution_arcsec, reference_year=2016)
        self.assertAlmostEqual(exp_bra.gdf.value.sum(), 1000)
        # compare number of data points:
        self.assertEqual(exp_bra.gdf.shape[0], exp_bra2.gdf.shape[0])
        self.assertEqual(exp_bra.gdf.shape[0], 3566)
        # check for double entries:
        self.assertEqual(len(exp_bra.gdf.geometry.unique()), len(exp_bra.gdf.geometry))
        self.assertEqual(len(exp_bra.gdf.geometry.unique()), 3566)

class TestGPWPopulation(unittest.TestCase):
    """Test gpw_population submodule"""

    def test_get_gpw_file_path_pass(self):
        """test method gpw_population.get_gpw_file_path"""
        gpw_version = CONFIG.exposures.litpop.gpw_population.gpw_version.int()
        try:
            path = gpw_population.get_gpw_file_path(gpw_version, 2020, verbatim=False)
            self.assertIn('gpw_v4_population', str(path))
        except FileExistsError as err:
            self.assertIn('lease download', err.args[0])
            self.skipTest('GPW input data for GPW v4.%i not found.' %(gpw_version))

    def test_load_gpw_pop_shape_pass(self):
        """test method gpw_population.load_gpw_pop_shape"""
        gpw_version = CONFIG.exposures.litpop.gpw_population.gpw_version.int()
        bounds = (8.41, 47.2, 8.70, 47.45) # (min_lon, max_lon, min_lat, max_lat)
        shape = Polygon([
            (bounds[0], bounds[3]),
            (bounds[2], bounds[3]),
            (bounds[2], bounds[1]),
            (bounds[0], bounds[1])
            ])
        try:
            data, meta, glb_transform = \
                gpw_population.load_gpw_pop_shape(shape, 2020, gpw_version,
                                                  verbatim=False)
            self.assertEqual(data.shape, (31, 36))
            self.assertAlmostEqual(meta['transform'][0], 0.00833333333333333)
            self.assertAlmostEqual(meta['transform'][0], glb_transform[0])

        except FileExistsError as err:
            self.assertIn('lease download', err.args[0])
            self.skipTest('GPW input data for GPW v4.%i not found.' %(gpw_version))

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestLitPopExposure)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestAdmin1))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Tag class
"""

import unittest

from climada.hazard.tag import Tag as TagHazard

class TestAppend(unittest.TestCase):
    """Test loading funcions from the Hazard class"""

    def test_append_right_pass(self):
        """Appends an other tag correctly."""
        tag1 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)
        self.assertEqual('TC', tag1.haz_type)

        tag2 = TagHazard('TC', 'file_name2.mat', 'dummy file 2')

        tag1.append(tag2)

        self.assertEqual(['file_name1.mat', 'file_name2.mat'], tag1.file_name)
        self.assertEqual(['dummy file 1', 'dummy file 2'], tag1.description)
        self.assertEqual('TC', tag1.haz_type)

    def test_append_wrong_pass(self):
        """Appends an other tag correctly."""
        tag1 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        tag2 = TagHazard('EQ', 'file_name2.mat', 'dummy file 2')
        with self.assertRaises(ValueError) as cm:
            tag1.append(tag2)
        self.assertIn("Hazards of different type can't be appended: TC != EQ.", str(cm.exception))

    def test_equal_same(self):
        """Appends an other tag correctly."""
        tag1 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        tag2 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        tag1.append(tag2)
        self.assertEqual(['file_name1.mat', 'file_name1.mat'], tag1.file_name)
        self.assertEqual(['dummy file 1', 'dummy file 1'], tag1.description)
        self.assertEqual('TC', tag1.haz_type)

    def test_append_empty(self):
        """Appends an other tag correctly."""
        tag1 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        tag2 = TagHazard()
        tag2.haz_type = 'TC'

        tag1.append(tag2)
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

        tag1 = TagHazard()
        tag1.haz_type = 'TC'
        tag2 = TagHazard('TC', 'file_name1.mat', 'dummy file 1')

        tag1.append(tag2)
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

class TestJoin(unittest.TestCase):
    """Test joining functions and string formation Tag class."""

    def test_one_str_pass(self):
        """Test __str__ method with one file"""
        tag = TagHazard('TC', 'file_name1.mat', 'dummy file 1')
        self.assertEqual(str(tag), ' Type: TC\n File: file_name1\n Description: dummy file 1')

    def test_teo_str_pass(self):
        """Test __str__ method with one file"""
        tag1 = TagHazard('TC', 'file1.mat', 'desc1')
        tag2 = TagHazard('TC', 'file2.xls', 'desc2')
        tag1.append(tag2)
        self.assertEqual(str(tag1),
                         ' Type: TC\n File: file1 + file2\n Description: desc1 + desc2')

    def test_join_names_pass(self):
        """Test join_file_names function."""
        tag1 = TagHazard('TC', 'file1', 'desc1')
        tag2 = TagHazard('TC', 'file2', 'desc2')
        tag1.append(tag2)
        join_name = tag1.join_file_names()
        self.assertEqual('file1 + file2', join_name)

    def test_join_descr_pass(self):
        """Test join_descriptions function."""
        tag1 = TagHazard('TC', 'file1', 'desc1')
        tag2 = TagHazard('TC', 'file2', 'desc2')
        tag1.append(tag2)
        join_desc = tag1.join_descriptions()
        self.assertEqual('desc1 + desc2', join_desc)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestAppend)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestJoin))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Tests on Black marble.
"""

import unittest
import numpy as np

from shapely.geometry import Polygon

from climada.entity.exposures.litpop import nightlight as nl_utils

from climada.util.constants import SYSTEM_DIR

def init_test_shape():
    bounds = (14.18, 35.78, 14.58, 36.09)
    # (min_lon, max_lon, min_lat, max_lat)

    return bounds, Polygon([
        (bounds[0], bounds[3]),
        (bounds[2], bounds[3]),
        (bounds[2], bounds[1]),
        (bounds[0], bounds[1])
        ])

class TestNightlight(unittest.TestCase):
    """Test litpop.nightlight"""

    def test_load_nasa_nl_shape_single_tile_pass(self):
        """load_nasa_nl_shape_single_tile pass"""
        bounds, shape = init_test_shape()

    def test_load_nasa_nl_2016_shape_pass(self):
        """load_nasa_nl_shape_single_tile pass"""
        bounds, shape = init_test_shape()
        nl_utils.load_nasa_nl_shape(shape, 2016, data_dir=SYSTEM_DIR, dtype=None)

        # data, meta = nl_utils.load_nasa_nl_shape_single_tile(shape, path, layer=0):


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestNightlight)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Tag class.
"""

__all__ = ['Tag']

class Tag():
    """Source data tag for Exposures, DiscRates, ImpactFuncSet, MeasureSet.

    Attributes
    ----------
    file_name : str
        name of the source file
    description : str
        description of the data
    """

    def __init__(self, file_name='', description=''):
        """Initialize values.

        Parameters
        ----------
        file_name : str, optional
            file name to read
        description : str, optional
            description of the data
        """
        self.file_name = str(file_name)
        self.description = description

    def append(self, tag):
        """Append input Tag instance information to current Tag."""
        # add file name if not present in tag
        if self.file_name == '':
            self.file_name = tag.file_name
            self.description = tag.description
        elif tag.file_name == '':
            return
        else:
            if tag.file_name not in self.file_name:
                self.file_name += ' + ' + tag.file_name
            if tag.description not in self.description:
                self.description += ' + ' + tag.description

    def __str__(self):
        return ' File: ' + self.file_name + '\n Description: ' + \
            self.description

    __repr__ = __str__



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Functions to merge EMDAT damages to hazard events.
"""
import logging
import pickle
from datetime import datetime
from pathlib import Path
import pandas as pd
import numpy as np
from cartopy.io import shapereader

from climada.util.finance import gdp
from climada.util.constants import DEF_CRS
import climada.util.coordinates as u_coord
from climada.engine import Impact
from climada.entity.tag import Tag
from climada.hazard.tag import Tag as TagHaz

LOGGER = logging.getLogger(__name__)

PERIL_SUBTYPE_MATCH_DICT = dict(TC=['Tropical cyclone'],
                                FL=['Coastal flood'],
                                EQ=['Ground movement', 'Earthquake'],
                                RF=['Riverine flood', 'Flood'],
                                WS=['Extra-tropical storm', 'Storm'],
                                DR=['Drought'],
                                LS=['Landslide'],
                                BF=['Forest fire', 'Wildfire', 'Land fire (Brush, Bush, Pastur']
                                )

PERIL_TYPE_MATCH_DICT = dict(DR=['Drought'],
                             EQ=['Earthquake'],
                             FL=['Flood'],
                             LS=['Landslide'],
                             VQ=['Volcanic activity'],
                             BF=['Wildfire'],
                             HW=['Extreme temperature']
                             )

VARNAMES_EMDAT = \
    {2018: {'Dis No': 'Disaster No.',
            'Disaster Type': 'Disaster type',
            'Disaster Subtype': 'Disaster subtype',
            'Event Name': 'Disaster name',
            'Country': 'Country',
            'ISO': 'ISO',
            'Location': 'Location',
            'Associated Dis': 'Associated disaster',
            'Associated Dis2': 'Associated disaster2',
            'Dis Mag Value': 'Magnitude value',
            'Dis Mag Scale': 'Magnitude scale',
            'Latitude': 'Latitude',
            'Longitude': 'Longitude',
            'Total Deaths': 'Total deaths',
            'Total Affected': 'Total affected',
            "Insured Damages ('000 US$)": "Insured losses ('000 US$)",
            "Total Damages ('000 US$)": "Total damage ('000 US$)"},
     2020: {'Dis No': 'Dis No',
            'Year': 'Year',
            'Seq': 'Seq',
            'Disaster Group': 'Disaster Group',
            'Disaster Subgroup': 'Disaster Subgroup',
            'Disaster Type': 'Disaster Type',
            'Disaster Subtype': 'Disaster Subtype',
            'Disaster Subsubtype': 'Disaster Subsubtype',
            'Event Name': 'Event Name',
            'Entry Criteria': 'Entry Criteria',
            'Country': 'Country',
            'ISO': 'ISO',
            'Region': 'Region',
            'Continent': 'Continent',
            'Location': 'Location',
            'Origin': 'Origin',
            'Associated Dis': 'Associated Dis',
            'Associated Dis2': 'Associated Dis2',
            'OFDA Response': 'OFDA Response',
            'Appeal': 'Appeal',
            'Declaration': 'Declaration',
            'Aid Contribution': 'Aid Contribution',
            'Dis Mag Value': 'Dis Mag Value',
            'Dis Mag Scale': 'Dis Mag Scale',
            'Latitude': 'Latitude',
            'Longitude': 'Longitude',
            'Local Time': 'Local Time',
            'River Basin': 'River Basin',
            'Start Year': 'Start Year',
            'Start Month': 'Start Month',
            'Start Day': 'Start Day',
            'End Year': 'End Year',
            'End Month': 'End Month',
            'End Day': 'End Day',
            'Total Deaths': 'Total Deaths',
            'No Injured': 'No Injured',
            'No Affected': 'No Affected',
            'No Homeless': 'No Homeless',
            'Total Affected': 'Total Affected',
            "Reconstruction Costs ('000 US$)": "Reconstruction Costs ('000 US$)",
            "Insured Damages ('000 US$)": "Insured Damages ('000 US$)",
            "Total Damages ('000 US$)": "Total Damages ('000 US$)",
            'CPI': 'CPI'}}


def assign_hazard_to_emdat(certainty_level, intensity_path_haz, names_path_haz,
                           reg_id_path_haz, date_path_haz, emdat_data,
                           start_time, end_time, keep_checks=False):
    """assign_hazard_to_emdat: link EMdat event to hazard

    Parameters
    ----------
    certainty_level : str
        'high' or 'low'
    intensity_path_haz : sparse matrix
        with hazards as rows and grid points as cols,
        values only at location with impacts
    names_path_haz : str
        identifier for each hazard (i.e. IBtracID) (rows of the matrix)
    reg_id_path_haz : str
        ISO country ID of each grid point (cols of the matrix)
    date_path_haz : str
        start date of each hazard (rows of the matrix)
    emdat_data: pd.DataFrame
        dataframe with EMdat data
    start_time : str
        start date of events to be assigned 'yyyy-mm-dd'
    end_time : str
        end date of events to be assigned 'yyyy-mm-dd'
    keep_checks : bool, optional

    Returns
    -------
    pd.dataframe with EMdat entries linked to a hazard
    """
    # check valid certainty level
    certainty_levels = ['high', 'low']
    if certainty_level not in certainty_levels:
        raise ValueError("Invalid certainty level. Expected one of: %s" % certainty_levels)

    # prepare hazard set
    print("Start preparing hazard set")
    hit_countries = hit_country_per_hazard(intensity_path_haz, names_path_haz,
                                           reg_id_path_haz, date_path_haz)
    # prepare damage set

    # adjust emdat_data to the path!!
    print("Start preparing damage set")
    lookup = create_lookup(emdat_data, start_time, end_time, disaster_subtype='Tropical cyclone')
    # calculate possible hits
    print("Calculate possible hits")
    hit5 = emdat_possible_hit(lookup=lookup, hit_countries=hit_countries, delta_t=5)
    hit5_match = match_em_id(lookup=lookup, poss_hit=hit5)
    print("1/5")
    hit10 = emdat_possible_hit(lookup=lookup, hit_countries=hit_countries, delta_t=10)
    hit10_match = match_em_id(lookup=lookup, poss_hit=hit10)
    print("2/5")
    hit15 = emdat_possible_hit(lookup=lookup, hit_countries=hit_countries, delta_t=15)
    hit15_match = match_em_id(lookup=lookup, poss_hit=hit15)
    print("3/5")
    hit25 = emdat_possible_hit(lookup=lookup, hit_countries=hit_countries, delta_t=25)
    hit25_match = match_em_id(lookup=lookup, poss_hit=hit25)
    print("4/5")
    hit50 = emdat_possible_hit(lookup=lookup, hit_countries=hit_countries, delta_t=50)
    hit50_match = match_em_id(lookup=lookup, poss_hit=hit50)
    print("5/5")

    # assign only tracks with high certainty
    print("Assign tracks")
    if certainty_level == 'high':
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit50_match, level=1)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit15_match,
                                    possible_tracks_2=hit50_match, level=2)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit25_match,
                                    possible_tracks_2=hit50_match, level=3)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit25_match, level=4)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit15_match,
                                    possible_tracks_2=hit25_match, level=5)
    # assign all tracks
    elif certainty_level == 'low':
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit5_match,
                                    possible_tracks_2=hit50_match, level=1)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit50_match, level=2)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit15_match,
                                    possible_tracks_2=hit50_match, level=3)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit5_match,
                                    possible_tracks_2=hit25_match, level=4)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit25_match, level=5)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit15_match,
                                    possible_tracks_2=hit25_match, level=6)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit5_match,
                                    possible_tracks_2=hit15_match, level=7)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit15_match, level=8)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit5_match,
                                    possible_tracks_2=hit10_match, level=9)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit15_match,
                                    possible_tracks_2=hit15_match, level=10)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit10_match,
                                    possible_tracks_2=hit10_match, level=11)
        lookup = assign_track_to_em(lookup=lookup, possible_tracks_1=hit5_match,
                                    possible_tracks_2=hit5_match, level=12)
    if not keep_checks:
        lookup = lookup.drop(['Date_start_EM_ordinal', 'possible_track',
                              'possible_track_all'], axis=1)
    lookup.groupby('allocation_level').count()
    print('(%d/%s) tracks allocated' % (
        len(lookup[lookup.allocation_level.notnull()]), len(lookup)))
    return lookup


def hit_country_per_hazard(intensity_path, names_path, reg_id_path, date_path):
    """hit_country_per_hazard: create list of hit countries from hazard set

    Parameters
    ----------
    intensity_path : str
        Path to file containing sparse matrix with hazards as rows and grid points
        as cols, values only at location with impacts
    names_path : str
        Path to file with identifier for each hazard (i.e. IBtracID) (rows of the matrix)
    reg_id_path : str
        Path to file with ISO country ID of each grid point (cols of the matrix)
    date_path : str
        Path to file with start date of each hazard (rows of the matrix)

    Returns
    -------
        pd.DataFrame with all hit countries per hazard
    """
    with open(intensity_path, 'rb') as filef:
        inten = pickle.load(filef)
    with open(names_path, 'rb') as filef:
        names = pickle.load(filef)
    with open(reg_id_path, 'rb') as filef:
        reg_id = pickle.load(filef)
    with open(date_path, 'rb') as filef:
        date = pickle.load(filef)
    # loop over the tracks (over the rows of the intensity matrix)
    all_hits = []
    for track in range(0, len(names)):
        # select track
        tc_track = inten[track, ]
        # select only indices that are not zero
        hits = tc_track.nonzero()[1]
        # get the country of these indices and remove dublicates
        hits = list(set(reg_id[hits]))
        # append hit countries to list
        all_hits.append(hits)

    # create data frame for output
    hit_countries = pd.DataFrame(columns=['hit_country', 'Date_start', 'ibtracsID'])
    for track, _ in enumerate(names):
        # Check if track has hit any country else go to the next track
        if len(all_hits[track]) > 0:
            # loop over hit_country
            for hit in range(0, len(all_hits[track])):
                # Hit country ISO
                ctry_iso = u_coord.country_to_iso(all_hits[track][hit], "alpha3")
                # create entry for each country a hazard has hit
                hit_countries = hit_countries.append({'hit_country': ctry_iso,
                                                      'Date_start': date[track],
                                                      'ibtracsID': names[track]},
                                                     ignore_index=True)
    # retrun data frame with all hit countries per hazard
    return hit_countries

def create_lookup(emdat_data, start, end, disaster_subtype='Tropical cyclone'):
    """create_lookup: prepare a lookup table of EMdat events to which hazards can be assigned

        Parameters
        ----------
        emdat_data: pd.DataFrame
            with EMdat data
        start : str
            start date of events to be assigned 'yyyy-mm-dd'
        end : str
            end date of events to be assigned 'yyyy-mm-dd'
        disaster_subtype : str
            EMdat disaster subtype

        Returns
        -------
        pd.DataFrame
        """
    data = emdat_data[emdat_data['Disaster_subtype'] == disaster_subtype]
    lookup = pd.DataFrame(columns=['hit_country', 'Date_start_EM',
                                   'Date_start_EM_ordinal', 'Disaster_name',
                                   'EM_ID', 'ibtracsID', 'allocation_level',
                                   'possible_track', 'possible_track_all'])
    lookup.hit_country = data.ISO
    lookup.Date_start_EM = data.Date_start_clean
    lookup.Disaster_name = data.Disaster_name
    lookup.EM_ID = data.Disaster_No
    lookup = lookup.reset_index(drop=True)
    # create ordinals
    for i in range(0, len(data.Date_start_clean.values)):
        lookup.Date_start_EM_ordinal[i] = datetime.toordinal(
            datetime.strptime(lookup.Date_start_EM.values[i], '%Y-%m-%d'))
        # ordinals to numeric
    lookup.Date_start_EM_ordinal = pd.to_numeric(lookup.Date_start_EM_ordinal)
    # select time
    emdat_start = datetime.toordinal(datetime.strptime(start, '%Y-%m-%d'))
    emdat_end = datetime.toordinal(datetime.strptime(end, '%Y-%m-%d'))

    lookup = lookup[lookup.Date_start_EM_ordinal.values > emdat_start]
    lookup = lookup[lookup.Date_start_EM_ordinal.values < emdat_end]

    return lookup


# Function to relate EM disaster to IBtrack using hit countries and time
def emdat_possible_hit(lookup, hit_countries, delta_t):
    """relate EM disaster to hazard using hit countries and time

    Parameters
    ----------
    lookup : pd.DataFrame
        to relate EMdatID to hazard
    delta_t :
        max time difference of start of EMdat event and hazard
    hit_countries:


    Returns
    -------
    list with possible hits
    """
    # lookup: PD dataframe that relates EMdatID to an IBtracsID
    # tracks: processed IBtracks with info which track hit which country
    # delta_t: time difference of start of EMdat and IBrtacks
    possible_hit_all = []
    for i in range(0, len(lookup.EM_ID.values)):
        possible_hit = []
        country_tracks = hit_countries[
            hit_countries['hit_country'] == lookup.hit_country.values[i]]
        for j in range(0, len(country_tracks.Date_start.values)):
            if (lookup.Date_start_EM_ordinal.values[i] - country_tracks.Date_start.values[j]) < \
                delta_t and (lookup.Date_start_EM_ordinal.values[i] -
                             country_tracks.Date_start.values[j]) >= 0:
                possible_hit.append(country_tracks.ibtracsID.values[j])
        possible_hit_all.append(possible_hit)

    return possible_hit_all


# function to check if EM_ID has been assigned already
def match_em_id(lookup, poss_hit):
    """function to check if EM_ID has been assigned already and combine possible hits

        Parameters
        ----------
        lookup : pd.dataframe
            to relate EMdatID to hazard
        poss_hit : list
            with possible hits

        Returns
        -------
        list
            with all possible hits per EMdat ID
        """
    possible_hit_all = []
    for i in range(0, len(lookup.EM_ID.values)):
        possible_hit = []
        # lookup without line i
        #lookup_match = lookup.drop(i)
        lookup_match = lookup
        # Loop over check if EM dat ID is the same
        for i_match in range(0, len(lookup_match.EM_ID.values)):
            if lookup.EM_ID.values[i] == lookup_match.EM_ID.values[i_match]:
                possible_hit.append(poss_hit[i])
        possible_hit_all.append(possible_hit)
    return possible_hit_all


def assign_track_to_em(lookup, possible_tracks_1, possible_tracks_2, level):
    """function to assign a hazard to an EMdat event
        to get some confidene into the procedure, hazards get only assigned
        if there is no other hazard occuring at a bigger time interval in that country
        Thus a track of possible_tracks_1 gets only assigned if there are no other
        tracks in possible_tracks_2.
        The confidence can be expressed with a certainty level

        Parameters
        ----------
        lookup : pd.DataFrame
            to relate EMdatID to hazard
        possible_tracks_1 : list
            list of possible hits with smaller time horizon
        possible_tracks_2 : list
            list of possible hits with larger time horizon
        level : int
            level of confidence

        Returns
        -------
        pd.DataFrame
            lookup with assigend tracks and possible hits
    """

    for i, _ in enumerate(possible_tracks_1):
        if np.isnan(lookup.allocation_level.values[i]):
            number_emdat_id = len(possible_tracks_1[i])
            # print(number_emdat_id)
            for j in range(0, number_emdat_id):
                # check that number of possible track stays the same at given
                # time difference and that list is not empty
                if len(possible_tracks_1[i][j]) == len(possible_tracks_2[i][j]) == 1 \
                        and possible_tracks_1[i][j] != []:
                    # check that all tracks are the same
                    if all(possible_tracks_1[i][0] == possible_tracks_1[i][k]
                           for k in range(0, len(possible_tracks_1[i]))):
                        # check that track ID has not been assigned to that country already
                        ctry_lookup = lookup[lookup['hit_country'] == lookup.hit_country.values[i]]
                        if possible_tracks_1[i][0][0] not in ctry_lookup.ibtracsID.values:
                            lookup.ibtracsID.values[i] = possible_tracks_1[i][0][0]
                            lookup.allocation_level.values[i] = level
                elif possible_tracks_1[i][j] != []:
                    lookup.possible_track.values[i] = possible_tracks_1[i]
        else:
            lookup.possible_track_all.values[i] = possible_tracks_1[i]
    return lookup

def check_assigned_track(lookup, checkset):
    """compare lookup with assigned tracks to a set with checked sets

        Parameters
        ----------
        lookup: pd.DataFrame
            dataframe to relate EMdatID to hazard
        checkset: pd.DataFrame
            dataframe with already checked hazards

        Returns
        -------
        error scores
    """
    # merge checkset and lookup
    check = pd.merge(checkset, lookup[['hit_country', 'EM_ID', 'ibtracsID']],
                     on=['hit_country', 'EM_ID'])
    check_size = len(check.ibtracsID.values)
    # not assigned values
    not_assigned = check.ibtracsID.isnull().sum(axis=0)
    # correct assigned values
    correct = sum(check.ibtracsID.values == check.IBtracsID_checked.values)
    # wrongly assigned values
    wrong = len(check.ibtracsID.values) - not_assigned - correct
    print('%.1f%% tracks assigned correctly, %.1f%% wrongly, %.1f%% not assigned'
          % (correct / check_size * 100,
             wrong / check_size * 100,
             not_assigned / check_size * 100))


def clean_emdat_df(emdat_file, countries=None, hazard=None, year_range=None,
                   target_version=2020):
    """
    Get a clean and standardized DataFrame from EM-DAT-CSV-file
    (1) load EM-DAT data from CSV to DataFrame and remove header/footer,
    (2) handle version, clean up, and add columns, and
    (3) filter by country, hazard type and year range (if any given)

    Parameters
    ----------
    emdat_file : str, Path, or DataFrame
        Either string with full path to CSV-file or
        pandas.DataFrame loaded from EM-DAT CSV
    countries : list of str
        country ISO3-codes or names, e.g. ['JAM', 'CUB'].
        countries=None for all countries (default)
    hazard : list or str
        List of Disaster (sub-)type accordung EMDAT terminology, i.e.:
        Animal accident, Drought, Earthquake, Epidemic, Extreme temperature,
        Flood, Fog, Impact, Insect infestation, Landslide, Mass movement (dry),
        Storm, Volcanic activity, Wildfire;
        Coastal Flooding, Convective Storm, Riverine Flood, Tropical cyclone,
        Tsunami, etc.;
        OR CLIMADA hazard type abbreviations, e.g. TC, BF, etc.
    year_range : list or tuple
        Year range to be extracted, e.g. (2000, 2015);
        (only min and max are considered)
    target_version : int
        required EM-DAT data format version (i.e. year of download),
        changes naming of columns/variables (default: 2020)

    Returns
    -------
    df_data : pd.DataFrame
        DataFrame containing cleaned and filtered EM-DAT impact data
    """
    # (1) load EM-DAT data from CSV to DataFrame, skipping the header:
    if isinstance(emdat_file, (str, Path)):
        df_emdat = pd.read_csv(emdat_file, encoding="ISO-8859-1", header=0)
        counter = 0
        while not ('Country' in df_emdat.columns and 'ISO' in df_emdat.columns):
            counter += 1
            df_emdat = pd.read_csv(emdat_file, encoding="ISO-8859-1", header=counter)
            if counter == 10:
                break
        del counter
    elif isinstance(emdat_file, pd.DataFrame):
        df_emdat = emdat_file
    else:
        raise TypeError('emdat_file needs to be str or DataFrame')
    # drop rows with 9 or more NaN values (e.g. footer):
    df_emdat = df_emdat.dropna(thresh=9)

    # (2)  handle version, clean up, and add columns:
    # (2.1) identify underlying EMDAT version of csv:
    version = 2020
    for vers in list(VARNAMES_EMDAT.keys()):
        if len(df_emdat.columns) >= len(VARNAMES_EMDAT[vers]) and \
           all(item in list(df_emdat.columns) for item in VARNAMES_EMDAT[vers].values()):
            version = vers
    # (2.2) create new DataFrame df_data with column names as target version
    df_data = pd.DataFrame(index=df_emdat.index.values,
                           columns=VARNAMES_EMDAT[target_version].values())
    if 'Year' not in df_data.columns:  # make sure column "Year" exists
        df_data['Year'] = np.nan
    for _, col in enumerate(df_data.columns):  # loop over columns
        if col in VARNAMES_EMDAT[version]:
            df_data[col] = df_emdat[VARNAMES_EMDAT[version][col]]
        elif col in df_emdat.columns:
            df_data[col] = df_emdat[col]
        elif col == 'Year' and version <= 2018:
            years_list = list()
            for _, disaster_no in enumerate(df_emdat[VARNAMES_EMDAT[version]['Dis No']]):
                if isinstance(disaster_no, str):
                    years_list.append(int(disaster_no[0:4]))
                else:
                    years_list.append(np.nan)
            df_data[col] = years_list
    if version <= 2018 and target_version >= 2020:
        date_list = list()
        year_list = list()
        month_list = list()
        day_list = list()
        for year in list(df_data['Year']):
            if not np.isnan(year):
                date_list.append(datetime.strptime(str(year), '%Y'))
            else:
                date_list.append(datetime.strptime(str('0001'), '%Y'))
        boolean_warning = True
        for idx, datestr in enumerate(list(df_emdat['Start date'])):
            try:
                date_list[idx] = datetime.strptime(datestr[-7:], '%m/%Y')
            except ValueError:
                if boolean_warning:
                    LOGGER.warning('EM_DAT CSV contains invalid time formats')
                    boolean_warning = False
            try:
                date_list[idx] = datetime.strptime(datestr, '%d/%m/%Y')
            except ValueError:
                if boolean_warning:
                    LOGGER.warning('EM_DAT CSV contains invalid time formats')
                    boolean_warning = False
            day_list.append(date_list[idx].day)
            month_list.append(date_list[idx].month)
            year_list.append(date_list[idx].year)
        df_data['Start Month'] = np.array(month_list, dtype='int')
        df_data['Start Day'] = np.array(day_list, dtype='int')
        df_data['Start Year'] = np.array(year_list, dtype='int')
        for var in ['Disaster Subtype', 'Disaster Type', 'Country']:
            df_data[VARNAMES_EMDAT[target_version][var]].fillna('None', inplace=True)

    # (3) Filter by countries, year range, and disaster type
    # (3.1) Countries:
    if countries and isinstance(countries, str):
        countries = [countries]
    if countries and isinstance(countries, list):
        for idx, country in enumerate(countries):
            # convert countries to iso3 alpha code:
            countries[idx] = u_coord.country_to_iso(country, "alpha3")
        df_data = df_data[df_data['ISO'].isin(countries)].reset_index(drop=True)
    # (3.2) Year range:
    if year_range:
        for idx in df_data.index:
            if np.isnan(df_data.loc[0, 'Year']):
                df_data.loc[0, 'Year'] = \
                    df_data.loc[0, VARNAMES_EMDAT[target_version]['Start Year']]
        df_data = df_data[(df_data['Year'] >= min(year_range)) &
                          (df_data['Year'] <= max(year_range))]

    # (3.3) Disaster type:
    if hazard and isinstance(hazard, str):
        hazard = [hazard]
    if hazard and isinstance(hazard, list):
        disaster_types = list()
        disaster_subtypes = list()
        for idx, haz in enumerate(hazard):
            if haz in df_data[VARNAMES_EMDAT[target_version]['Disaster Type']].unique():
                disaster_types.append(haz)
            if haz in df_data[VARNAMES_EMDAT[target_version]['Disaster Subtype']].unique():
                disaster_subtypes.append(haz)
            if haz in PERIL_TYPE_MATCH_DICT.keys():
                disaster_types += PERIL_TYPE_MATCH_DICT[haz]
            if haz in PERIL_SUBTYPE_MATCH_DICT.keys():
                disaster_subtypes += PERIL_SUBTYPE_MATCH_DICT[haz]
        df_data = df_data[
            (df_data[VARNAMES_EMDAT[target_version]['Disaster Type']].isin(disaster_types)) |
            (df_data[VARNAMES_EMDAT[target_version]['Disaster Subtype']].isin(disaster_subtypes))]
    return df_data.reset_index(drop=True)

def emdat_countries_by_hazard(emdat_file_csv, hazard=None, year_range=None):
    """return list of all countries exposed to a chosen hazard type
    from EMDAT data as CSV.

    Parameters
    ----------
    emdat_file : str, Path, or DataFrame
        Either string with full path to CSV-file or
        pandas.DataFrame loaded from EM-DAT CSV
    hazard : list or str
        List of Disaster (sub-)type accordung EMDAT terminology, i.e.:
        Animal accident, Drought, Earthquake, Epidemic, Extreme temperature,
        Flood, Fog, Impact, Insect infestation, Landslide, Mass movement (dry),
        Storm, Volcanic activity, Wildfire;
        Coastal Flooding, Convective Storm, Riverine Flood, Tropical cyclone,
        Tsunami, etc.;
        OR CLIMADA hazard type abbreviations, e.g. TC, BF, etc.
    year_range : list or tuple
        Year range to be extracted, e.g. (2000, 2015);
        (only min and max are considered)

    Returns
    -------
    countries_iso3a : list
        List of ISO3-codes of countries impacted by the disaster (sub-)types
    countries_names : list
        List of names of countries impacted by the disaster (sub-)types
    """
    df_data = clean_emdat_df(emdat_file_csv, hazard=hazard, year_range=year_range)
    countries_iso3a = list(df_data.ISO.unique())
    countries_names = list()
    for iso3a in countries_iso3a:
        try:
            countries_names.append(u_coord.country_to_iso(iso3a, "name"))
        except LookupError:
            countries_names.append('NA')
    return countries_iso3a, countries_names


def scale_impact2refyear(impact_values, year_values, iso3a_values, reference_year=None):
    """Scale give impact values proportional to GDP to the according value in a reference year
    (for normalization of monetary values)

    Parameters
    ----------
    impact_values : list or array
        Impact values to be scaled.
    year_values : list or array
        Year of each impact (same length as impact_values)
    iso3a_values : list or array
        ISO3alpha code of country for each impact (same length as impact_values)

    Optional Parameters
    -------------------
    reference_year : int
        Impact is scaled proportional to GDP to the value of the reference year.
        No scaling for reference_year=None (default)
        """
    impact_values = np.array(impact_values)
    year_values = np.array(year_values)
    iso3a_values = np.array(iso3a_values)
    if reference_year and isinstance(reference_year, (int, float)):
        reference_year = int(reference_year)
        gdp_ref = dict()
        gdp_years = dict()
        for country in np.unique(iso3a_values):
            # get reference GDP value for each country:
            gdp_ref[country] = gdp(country, reference_year)[1]
            # get GDP value for each country and year:
            gdp_years[country] = dict()
            years_country = np.unique(year_values[iso3a_values == country])
            print(years_country)
            for year in years_country:
                gdp_years[country][year] = gdp(country, year)[1]
        # loop through each value and apply scaling:
        for idx, val in enumerate(impact_values):
            impact_values[idx] = val * gdp_ref[iso3a_values[idx]] / \
                gdp_years[iso3a_values[idx]][year_values[idx]]
        return list(impact_values)
    if not reference_year:
        return impact_values
    raise ValueError('Invalid reference_year')

def emdat_impact_yearlysum(emdat_file_csv, countries=None, hazard=None, year_range=None,
                           reference_year=None, imp_str="Total Damages ('000 US$)",
                           version=2020):
    """function to load EM-DAT data and sum impact per year

    Parameters
    ----------
     emdat_file_csv : str or DataFrame
        Either string with full path to CSV-file or
        pandas.DataFrame loaded from EM-DAT CSV
    countries : list of str
        country ISO3-codes or names, e.g. ['JAM', 'CUB'].
        countries=None for all countries (default)
    hazard : list or str
        List of Disaster (sub-)type accordung EMDAT terminology, i.e.:
        Animal accident, Drought, Earthquake, Epidemic, Extreme temperature,
        Flood, Fog, Impact, Insect infestation, Landslide, Mass movement (dry),
        Storm, Volcanic activity, Wildfire;
        Coastal Flooding, Convective Storm, Riverine Flood, Tropical cyclone,
        Tsunami, etc.;
        OR CLIMADA hazard type abbreviations, e.g. TC, BF, etc.
    year_range : list or tuple
        Year range to be extracted, e.g. (2000, 2015);
        (only min and max are considered)
    version : int
        required EM-DAT data format version (i.e. year of download),
        changes naming of columns/variables (default: 2020)

    Returns
    -------
    out : pd.DataFrame
        DataFrame with summed impact and scaled impact per
        year and country.
    """
    imp_str = VARNAMES_EMDAT[version][imp_str]
    df_data = clean_emdat_df(emdat_file_csv, countries=countries, hazard=hazard,
                             year_range=year_range, target_version=version)

    df_data[imp_str + " scaled"] = scale_impact2refyear(df_data[imp_str].values,
                                                        df_data.Year.values, df_data.ISO.values,
                                                        reference_year=reference_year)
    out = pd.DataFrame(columns=['ISO', 'region_id', 'year', 'impact',
                                'impact_scaled', 'reference_year'])
    for country in df_data.ISO.unique():
        country = u_coord.country_to_iso(country, "alpha3")
        if not df_data.loc[df_data.ISO == country].size:
            continue
        all_years = np.arange(min(df_data.Year), max(df_data.Year) + 1)
        data_out = pd.DataFrame(index=np.arange(0, len(all_years)),
                                columns=out.columns)
        df_country = df_data.loc[df_data.ISO == country]
        for cnt, year in enumerate(all_years):
            data_out.loc[cnt, 'year'] = year
            data_out.loc[cnt, 'reference_year'] = reference_year
            data_out.loc[cnt, 'ISO'] = country
            data_out.loc[cnt, 'region_id'] = u_coord.country_to_iso(country, "numeric")
            data_out.loc[cnt, 'impact'] = \
                np.nansum(df_country[df_country.Year.isin([year])][imp_str])
            data_out.loc[cnt, 'impact_scaled'] = \
                np.nansum(df_country[df_country.Year.isin([year])][imp_str + " scaled"])
            if '000 US' in imp_str:  # EM-DAT damages provided in '000 USD
                data_out.loc[cnt, 'impact'] = data_out.loc[cnt, 'impact'] * 1e3
                data_out.loc[cnt, 'impact_scaled'] = data_out.loc[cnt, 'impact_scaled'] * 1e3
        out = out.append(data_out)
    out = out.reset_index(drop=True)
    return out

def emdat_impact_event(emdat_file_csv, countries=None, hazard=None, year_range=None,
                       reference_year=None, imp_str="Total Damages ('000 US$)",
                       version=2020):
    """function to load EM-DAT data return impact per event

    Parameters
    ----------
     emdat_file_csv : str or DataFrame
        Either string with full path to CSV-file or
        pandas.DataFrame loaded from EM-DAT CSV
    countries : list of str
        country ISO3-codes or names, e.g. ['JAM', 'CUB'].
        default: countries=None for all countries
    hazard : list or str
        List of Disaster (sub-)type accordung EMDAT terminology, i.e.:
        Animal accident, Drought, Earthquake, Epidemic, Extreme temperature,
        Flood, Fog, Impact, Insect infestation, Landslide, Mass movement (dry),
        Storm, Volcanic activity, Wildfire;
        Coastal Flooding, Convective Storm, Riverine Flood, Tropical cyclone,
        Tsunami, etc.;
        OR CLIMADA hazard type abbreviations, e.g. TC, BF, etc.
    year_range : list or tuple
        Year range to be extracted, e.g. (2000, 2015);
        (only min and max are considered)
    reference_year : int reference year of exposures. Impact is scaled
        proportional to GDP to the value of the reference year. Default: No scaling
        for 0
    imp_str : str
        Column name of impact metric in EMDAT CSV,
        default = "Total Damages ('000 US$)"
    version : int
        EM-DAT version to take variable/column names from (defaul: 2020)

    Returns
    -------
    out : pd.DataFrame
        EMDAT DataFrame with new columns "year",
        "region_id", and "impact" and +impact_scaled" total impact per event with
        same unit as chosen impact, but multiplied by 1000 if impact is given
        as 1000 US$ (e.g. imp_str="Total Damages ('000 US$) scaled").
    """
    imp_str = VARNAMES_EMDAT[version][imp_str]
    df_data = clean_emdat_df(emdat_file_csv, hazard=hazard, year_range=year_range,
                             countries=countries, target_version=version)
    df_data['year'] = df_data['Year']
    df_data['reference_year'] = reference_year
    df_data['impact'] = df_data[imp_str]
    df_data['impact_scaled'] = scale_impact2refyear(df_data[imp_str].values, df_data.Year.values,
                                                    df_data.ISO.values,
                                                    reference_year=reference_year)
    df_data['region_id'] = np.nan
    for country in df_data.ISO.unique():
        try:
            df_data.loc[df_data.ISO == country, 'region_id'] = \
                u_coord.country_to_iso(country, "numeric")
        except LookupError:
            LOGGER.warning('ISO3alpha code not found in iso_country: %s', country)
    if '000 US' in imp_str:
        df_data['impact'] *= 1e3
        df_data['impact_scaled'] *= 1e3
    return df_data.reset_index(drop=True)

def emdat_to_impact(emdat_file_csv, hazard_type_climada, year_range=None, countries=None,
                    hazard_type_emdat=None,
                    reference_year=None, imp_str="Total Damages"):
    """function to load EM-DAT data return impact per event

    Parameters
    ----------
     emdat_file_csv : str or pd.DataFrame
        Either string with full path to CSV-file or
        pandas.DataFrame loaded from EM-DAT CSV
    countries : list of str
        country ISO3-codes or names, e.g. ['JAM', 'CUB'].
        default: countries=None for all countries
    hazard_type_climada : list or str
        List of Disaster (sub-)type accordung EMDAT terminology, i.e.:
        Animal accident, Drought, Earthquake, Epidemic, Extreme temperature,
        Flood, Fog, Impact, Insect infestation, Landslide, Mass movement (dry),
        Storm, Volcanic activity, Wildfire;
        Coastal Flooding, Convective Storm, Riverine Flood, Tropical cyclone,
        Tsunami, etc.;
        OR CLIMADA hazard type abbreviations, e.g. TC, BF, etc.
    year_range : list or tuple
        Year range to be extracted, e.g. (2000, 2015);
        (only min and max are considered)
    reference_year : int reference year of exposures. Impact is scaled
        proportional to GDP to the value of the reference year. Default: No scaling
        for 0
    imp_str : str
        Column name of impact metric in EMDAT CSV,
        default = "Total Damages ('000 US$)"

    Returns
    -------
    impact_instance : instance of climada.engine.Impact
    impact object of same format as output from CLIMADA
    impact computation.
    Values scaled with GDP to reference_year if reference_year is given.
    i.e. current US$ for imp_str="Total Damages ('000 US$) scaled" (factor 1000 is applied)
    impact_instance.eai_exp holds expected annual impact for each country.
    impact_instance.coord_exp holds rough central coordinates for each country.
    countries (list): ISO3-codes of countries in same order as in impact_instance.eai_exp
    """
    if "Total Damages" in imp_str:
        imp_str = "Total Damages ('000 US$)"
    elif "Insured Damages" in imp_str:
        imp_str = "Insured Damages ('000 US$)"
    elif "Reconstruction Costs" in imp_str:
        imp_str = "Reconstruction Costs ('000 US$)"
    imp_str = VARNAMES_EMDAT[max(VARNAMES_EMDAT.keys())][imp_str]
    if not hazard_type_emdat:
        hazard_type_emdat = [hazard_type_climada]
    if reference_year == 0:
        reference_year = None
    # Inititate Impact-instance:
    impact_instance = Impact()

    impact_instance.tag = dict()
    impact_instance.tag['haz'] = TagHaz(haz_type=hazard_type_climada,
                                        file_name=emdat_file_csv,
                                        description='EM-DAT impact, direct import')
    impact_instance.tag['exp'] = Tag(file_name=emdat_file_csv,
                                     description='EM-DAT impact, direct import')
    impact_instance.tag['impf_set'] = Tag(file_name=None, description=None)


    # Load EM-DAT impact data by event:
    em_data = emdat_impact_event(emdat_file_csv, countries=countries, hazard=hazard_type_emdat,
                                 year_range=year_range, reference_year=reference_year,
                                 imp_str=imp_str, version=max(VARNAMES_EMDAT.keys()))

    if isinstance(countries, str):
        countries = [countries]
    elif not countries:
        countries = emdat_countries_by_hazard(emdat_file_csv, year_range=year_range,
                                              hazard=hazard_type_emdat)[0]

    if em_data.empty:
        return impact_instance, countries
    impact_instance.event_id = np.array(em_data.index, int)
    impact_instance.event_name = list(
        em_data[VARNAMES_EMDAT[max(VARNAMES_EMDAT.keys())]['Dis No']])

    date_list = list()
    for year in list(em_data['Year']):
        date_list.append(datetime.toordinal(datetime.strptime(str(year), '%Y')))
    if 'Start Year' in em_data.columns and 'Start Month' in em_data.columns \
            and 'Start Day' in em_data.columns:
        idx = 0
        for year, month, day in zip(em_data['Start Year'], em_data['Start Month'],
                                    em_data['Start Day']):
            if np.isnan(year):
                idx += 1
                continue
            if np.isnan(month):
                month = 1
            if np.isnan(day):
                day = 1
            date_list[idx] = datetime.toordinal(datetime.strptime(
                '%02i/%02i/%04i' % (day, month, year), '%d/%m/%Y'))
            idx += 1
    impact_instance.date = np.array(date_list, int)
    impact_instance.crs = DEF_CRS

    if not reference_year:
        impact_instance.at_event = np.array(em_data["impact"])
    else:
        impact_instance.at_event = np.array(em_data["impact_scaled"])
    impact_instance.at_event[np.isnan(impact_instance.at_event)] = 0
    if not year_range:
        year_range = [em_data['Year'].min(), em_data['Year'].max()]
    impact_instance.frequency = np.ones(em_data.shape[0]) / (1 + np.diff(year_range))
    impact_instance.tot_value = 0
    impact_instance.aai_agg = np.nansum(impact_instance.at_event * impact_instance.frequency)
    impact_instance.unit = 'USD'
    impact_instance.imp_mat = []

    # init rough exposure with central point per country
    shp = shapereader.natural_earth(resolution='110m',
                                    category='cultural',
                                    name='admin_0_countries')
    shp = shapereader.Reader(shp)
    countries_reg_id = list()
    countries_lat = list()
    countries_lon = list()
    impact_instance.eai_exp = np.zeros(len(countries))  # empty: damage at exposure
    for idx, cntry in enumerate(countries):
        try:
            cntry = u_coord.country_to_iso(cntry, "alpha3")
        except LookupError:
            LOGGER.warning('Country not found in iso_country: %s', cntry)
        cntry_boolean = False
        for rec in shp.records():
            if rec.attributes['ADM0_A3'].casefold() == cntry.casefold():
                bbox = rec.geometry.bounds
                cntry_boolean = True
                break
        if cntry_boolean:
            countries_lat.append(np.mean([bbox[1], bbox[3]]))
            countries_lon.append(np.mean([bbox[0], bbox[2]]))
        else:
            countries_lat.append(np.nan)
            countries_lon.append(np.nan)
        try:
            countries_reg_id.append(u_coord.country_to_iso(cntry, "numeric"))
        except LookupError:
            countries_reg_id.append(0)
        df_tmp = em_data[em_data[VARNAMES_EMDAT[
            max(VARNAMES_EMDAT.keys())]['ISO']].str.contains(cntry)]
        if not reference_year:
            impact_instance.eai_exp[idx] = sum(np.array(df_tmp["impact"]) *
                                               impact_instance.frequency[0])
        else:
            impact_instance.eai_exp[idx] = sum(np.array(df_tmp["impact_scaled"]) *
                                               impact_instance.frequency[0])

    impact_instance.coord_exp = np.stack([countries_lat, countries_lon], axis=1)
    return impact_instance, countries
